import unittest

from pydistsim.demo_algorithms.niculescu2003.dvhop import DVHop
from pydistsim.demo_algorithms.niculescu2003.trilaterate import Trilaterate
from pydistsim.network import NetworkGenerator
from pydistsim.network.sensor import TruePosSensor
from pydistsim.simulation import Simulation


class TestNiculescu2003(unittest.TestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100, directed=False)
        self.net = net_gen.generate_random_network()
        self.algorithms = (
            (DVHop, {"truePositionKey": "tp", "hopsizeKey": "hs", "dataKey": "I"}),
            (
                Trilaterate,
                {
                    "truePositionKey": "tp",
                    "hopsizeKey": "hs",
                    "positionKey": "pos",
                    "dataKey": "I",
                },
            ),
        )
        for node in self.net.nodes_sorted()[:10]:
            node.compositeSensor = (TruePosSensor,)

    def test_niculescu2003_sim(self):
        """Test niculescu2003 default simulation."""
        sim = Simulation(self.net)
        sim.algorithms = self.algorithms
        sim.run(100_000)
        for node in self.net.nodes():
            self.assertTrue(len(node.memory.get("pos", [None, None])) == 2 or "tp" in node.memory)



load_subconfig("ipython_config.py", profile="pydistsim")
# Configuration file for ipython-notebook.

c = get_config()

# ------------------------------------------------------------------------------
# NotebookApp configuration
# ------------------------------------------------------------------------------

# NotebookApp will inherit config from: BaseIPythonApplication, Application

# The IPython profile to use.
# c.NotebookApp.profile = u'default'

# The url for MathJax.js.
# c.NotebookApp.mathjax_url = ''

# The IP address the notebook server will listen on.
# c.NotebookApp.ip = '127.0.0.1'

# The base URL for the notebook server
# c.NotebookApp.base_project_url = '/'

# Create a massive crash report when IPython encounters what may be an internal
# error.  The default is to append a short message to the usual traceback
# c.NotebookApp.verbose_crash = False

# The number of additional ports to try if the specified port is not available.
# c.NotebookApp.port_retries = 50

# Whether to install the default config files into the profile dir. If a new
# profile is being created, and IPython contains config files for that profile,
# then they will be staged into the new directory.  Otherwise, default config
# files will be automatically generated.
# c.NotebookApp.copy_config_files = False

# The base URL for the kernel server
# c.NotebookApp.base_kernel_url = '/'

# The port the notebook server will listen on.
# c.NotebookApp.port = 8888

# Whether to overwrite existing config files when copying
# c.NotebookApp.overwrite = False

# Whether to prevent editing/execution of notebooks.
# c.NotebookApp.read_only = False

# Whether to enable MathJax for typesetting math/TeX
#
# MathJax is the javascript library IPython uses to render math/LaTeX. It is
# very large, so you may want to disable it if you have a slow internet
# connection, or for offline use of the notebook.
#
# When disabled, equations etc. will appear as their untransformed TeX source.
# c.NotebookApp.enable_mathjax = True

# Whether to open in a browser after starting. The specific browser used is
# platform dependent and determined by the python standard library `webbrowser`
# module, unless it is overridden using the --browser (NotebookApp.browser)
# configuration option.
# c.NotebookApp.open_browser = True

# The full path to an SSL/TLS certificate file.
# c.NotebookApp.certfile = u''

# The hostname for the websocket server.
# c.NotebookApp.websocket_host = ''

# The name of the IPython directory. This directory is used for logging
# configuration (through profiles), history storage, etc. The default is usually
# $HOME/.ipython. This options can also be specified through the environment
# variable IPYTHONDIR.
# c.NotebookApp.ipython_dir = u''

# Set the log level by value or name.
# c.NotebookApp.log_level = 20

# Hashed password to use for web authentication.
#
# To generate, type in a python/IPython shell:
#
#   from IPython.lib import passwd; passwd()
#
# The string should be of the form type:salt:hashed-password.
# c.NotebookApp.password = u''

# The Logging format template
# c.NotebookApp.log_format = '[%(name)s] %(message)s'

# The full path to a private key file for usage with SSL/TLS.
# c.NotebookApp.keyfile = u''

# Supply overrides for the tornado.web.Application that the IPython notebook
# uses.
# c.NotebookApp.webapp_settings = {}

# Specify what command to use to invoke a web browser when opening the notebook.
# If not specified, the default browser will be determined by the `webbrowser`
# standard library module, which allows setting of the BROWSER environment
# variable to override it.
# c.NotebookApp.browser = u''

# ------------------------------------------------------------------------------
# IPKernelApp configuration
# ------------------------------------------------------------------------------

# IPython: an enhanced interactive Python shell.

# IPKernelApp will inherit config from: KernelApp, BaseIPythonApplication,
# Application, InteractiveShellApp

# The importstring for the DisplayHook factory
# c.IPKernelApp.displayhook_class = 'IPython.zmq.displayhook.ZMQDisplayHook'

# Set the IP or interface on which the kernel will listen.
# c.IPKernelApp.ip = '127.0.0.1'

#
# c.IPKernelApp.parent_appname = u''

# Create a massive crash report when IPython encounters what may be an internal
# error.  The default is to append a short message to the usual traceback
# c.IPKernelApp.verbose_crash = False

# Run the module as a script.
# c.IPKernelApp.module_to_run = ''

# set the shell (ROUTER) port [default: random]
# c.IPKernelApp.shell_port = 0

# Whether to overwrite existing config files when copying
# c.IPKernelApp.overwrite = False

# Execute the given command string.
# c.IPKernelApp.code_to_run = ''

# set the stdin (DEALER) port [default: random]
# c.IPKernelApp.stdin_port = 0

# Set the log level by value or name.
# c.IPKernelApp.log_level = 30

# lines of code to run at IPython startup.
# c.IPKernelApp.exec_lines = []

# The importstring for the OutStream factory
# c.IPKernelApp.outstream_class = 'IPython.zmq.iostream.OutStream'

# Whether to create profile dir if it doesn't exist
# c.IPKernelApp.auto_create = False

# set the heartbeat port [default: random]
# c.IPKernelApp.hb_port = 0

# redirect stdout to the null device
# c.IPKernelApp.no_stdout = False

# dotted module name of an IPython extension to load.
# c.IPKernelApp.extra_extension = ''

# A file to be run
# c.IPKernelApp.file_to_run = ''

# The IPython profile to use.
# c.IPKernelApp.profile = u'default'

# Pre-load matplotlib and numpy for interactive use, selecting a particular
# matplotlib backend and loop integration.
c.IPKernelApp.pylab = "inline"

# kill this process if its parent dies.  On Windows, the argument specifies the
# HANDLE of the parent process, otherwise it is simply boolean.
# c.IPKernelApp.parent = 0

# JSON file in which to store connection info [default: kernel-<pid>.json]
#
# This file will contain the IP, ports, and authentication key needed to connect
# clients to this kernel. By default, this file will be created in the security-
# dir of the current profile, but can be specified by absolute path.
# c.IPKernelApp.connection_file = ''

# If true, an 'import *' is done from numpy and pylab, when using pylab
# c.IPKernelApp.pylab_import_all = True

# The name of the IPython directory. This directory is used for logging
# configuration (through profiles), history storage, etc. The default is usually
# $HOME/.ipython. This options can also be specified through the environment
# variable IPYTHONDIR.
# c.IPKernelApp.ipython_dir = u''

# ONLY USED ON WINDOWS Interrupt this process when the parent is signalled.
# c.IPKernelApp.interrupt = 0

# Whether to install the default config files into the profile dir. If a new
# profile is being created, and IPython contains config files for that profile,
# then they will be staged into the new directory.  Otherwise, default config
# files will be automatically generated.
# c.IPKernelApp.copy_config_files = False

# List of files to run at IPython startup.
# c.IPKernelApp.exec_files = []

# Enable GUI event loop integration ('qt', 'wx', 'gtk', 'glut', 'pyglet',
# 'osx').
# c.IPKernelApp.gui = None

# A list of dotted module names of IPython extensions to load.
# c.IPKernelApp.extensions = []

# redirect stderr to the null device
# c.IPKernelApp.no_stderr = False

# The Logging format template
# c.IPKernelApp.log_format = '[%(name)s] %(message)s'

# set the iopub (PUB) port [default: random]
# c.IPKernelApp.iopub_port = 0

# ------------------------------------------------------------------------------
# ZMQInteractiveShell configuration
# ------------------------------------------------------------------------------

# A subclass of InteractiveShell for ZMQ.

# ZMQInteractiveShell will inherit config from: InteractiveShell

# Use colors for displaying information about objects. Because this information
# is passed through a pager (like 'less'), and some pagers get confused with
# color codes, this capability can be turned off.
# c.ZMQInteractiveShell.color_info = True

#
# c.ZMQInteractiveShell.history_length = 10000

# Don't call post-execute functions that have failed in the past.
# c.ZMQInteractiveShell.disable_failing_post_execute = False

# Show rewritten input, e.g. for autocall.
# c.ZMQInteractiveShell.show_rewritten_input = True

# Set the color scheme (NoColor, Linux, or LightBG).
# c.ZMQInteractiveShell.colors = 'Linux'

#
# c.ZMQInteractiveShell.separate_in = '\n'

# Deprecated, use PromptManager.in2_template
# c.ZMQInteractiveShell.prompt_in2 = '   .\\D.: '

#
# c.ZMQInteractiveShell.separate_out = ''

# Deprecated, use PromptManager.in_template
# c.ZMQInteractiveShell.prompt_in1 = 'In [\\#]: '

# Enable deep (recursive) reloading by default. IPython can use the deep_reload
# module which reloads changes in modules recursively (it replaces the reload()
# function, so you don't need to change anything to use it). deep_reload()
# forces a full reload of modules whose code may have changed, which the default
# reload() function does not.  When deep_reload is off, IPython will use the
# normal reload(), but deep_reload will still be available as dreload().
# c.ZMQInteractiveShell.deep_reload = False

# Make IPython automatically call any callable object even if you didn't type
# explicit parentheses. For example, 'str 43' becomes 'str(43)' automatically.
# The value can be '0' to disable the feature, '1' for 'smart' autocall, where
# it is not applied if there are no more arguments on the line, and '2' for
# 'full' autocall, where all callable objects are automatically called (even if
# no arguments are present).
# c.ZMQInteractiveShell.autocall = 0

#
# c.ZMQInteractiveShell.separate_out2 = ''

# Deprecated, use PromptManager.justify
# c.ZMQInteractiveShell.prompts_pad_left = True

#
# c.ZMQInteractiveShell.readline_parse_and_bind = ['tab: complete', '"\\C-l": clear-screen', 'set show-all-if-ambiguous on', '"\\C-o": tab-insert', '"\\C-r": reverse-search-history', '"\\C-s": forward-search-history', '"\\C-p": history-search-backward', '"\\C-n": history-search-forward', '"\\e[A": history-search-backward', '"\\e[B": history-search-forward', '"\\C-k": kill-line', '"\\C-u": unix-line-discard']

# Enable magic commands to be called without the leading %.
# c.ZMQInteractiveShell.automagic = True

#
# c.ZMQInteractiveShell.debug = False

#
# c.ZMQInteractiveShell.object_info_string_level = 0

#
# c.ZMQInteractiveShell.ipython_dir = ''

#
# c.ZMQInteractiveShell.readline_remove_delims = '-/~'

# Start logging to the default log file.
# c.ZMQInteractiveShell.logstart = False

# The name of the logfile to use.
# c.ZMQInteractiveShell.logfile = ''

#
# c.ZMQInteractiveShell.wildcards_case_sensitive = True

# Save multi-line entries as one entry in readline history
# c.ZMQInteractiveShell.multiline_history = False

# Start logging to the given file in append mode.
# c.ZMQInteractiveShell.logappend = ''

#
# c.ZMQInteractiveShell.xmode = 'Context'

#
# c.ZMQInteractiveShell.quiet = False

# Deprecated, use PromptManager.out_template
# c.ZMQInteractiveShell.prompt_out = 'Out[\\#]: '

# Set the size of the output cache.  The default is 1000, you can change it
# permanently in your config file.  Setting it to 0 completely disables the
# caching system, and the minimum value accepted is 20 (if you provide a value
# less than 20, it is reset to 0 and a warning is issued).  This limit is
# defined because otherwise you'll spend more time re-flushing a too small cache
# than working
# c.ZMQInteractiveShell.cache_size = 1000

# 'all', 'last', 'last_expr' or 'none', specifying which nodes should be run
# interactively (displaying output from expressions).
# c.ZMQInteractiveShell.ast_node_interactivity = 'last_expr'

# Automatically call the pdb debugger after every exception.
# c.ZMQInteractiveShell.pdb = False

# ------------------------------------------------------------------------------
# ProfileDir configuration
# ------------------------------------------------------------------------------

# An object to manage the profile directory and its resources.
#
# The profile directory is used by all IPython applications, to manage
# configuration, logging and security.
#
# This object knows how to find, create and manage these directories. This
# should be used by any code that wants to handle profiles.

# Set the profile location directly. This overrides the logic used by the
# `profile` option.
# c.ProfileDir.location = u''

# ------------------------------------------------------------------------------
# Session configuration
# ------------------------------------------------------------------------------

# Object for handling serialization and sending of messages.
#
# The Session object handles building messages and sending them with ZMQ sockets
# or ZMQStream objects.  Objects can communicate with each other over the
# network via Session objects, and only need to work with the dict-based IPython
# message spec. The Session will handle serialization/deserialization, security,
# and metadata.
#
# Sessions support configurable serialiization via packer/unpacker traits, and
# signing with HMAC digests via the key/keyfile traits.
#
# Parameters ----------
#
# debug : bool
#     whether to trigger extra debugging statements
# packer/unpacker : str : 'json', 'pickle' or import_string
#     importstrings for methods to serialize message parts.  If just
#     'json' or 'pickle', predefined JSON and pickle packers will be used.
#     Otherwise, the entire importstring must be used.
#
#     The functions must accept at least valid JSON input, and output *bytes*.
#
#     For example, to use msgpack:
#     packer = 'msgpack.packb', unpacker='msgpack.unpackb'
# pack/unpack : callables
#     You can also set the pack/unpack callables for serialization directly.
# session : bytes
#     the ID of this Session object.  The default is to generate a new UUID.
# username : unicode
#     username added to message headers.  The default is to ask the OS.
# key : bytes
#     The key used to initialize an HMAC signature.  If unset, messages
#     will not be signed or checked.
# keyfile : filepath
#     The file containing a key.  If this is set, `key` will be initialized
#     to the contents of the file.

# Username for the Session. Default is your system username.
# c.Session.username = u'username'

# The name of the packer for serializing messages. Should be one of 'json',
# 'pickle', or an import name for a custom callable serializer.
# c.Session.packer = 'json'

# The UUID identifying this session.
# c.Session.session = u''

# execution key, for extra authentication.
# c.Session.key = ''

# Debug output in the Session
# c.Session.debug = False

# The name of the unpacker for unserializing messages. Only used with custom
# functions for `packer`.
# c.Session.unpacker = 'json'

# path to file containing execution key.
# c.Session.keyfile = ''

# ------------------------------------------------------------------------------
# InlineBackend configuration
# ------------------------------------------------------------------------------

# An object to store configuration of the inline backend.

# The image format for figures with the inline backend.
# c.InlineBackend.figure_format = 'png'

# Close all figures at the end of each cell.
#
# When True, ensures that each cell starts with no active figures, but it also
# means that one must keep track of references in order to edit or redraw
# figures in subsequent cells. This mode is ideal for the notebook, where
# residual plots from other cells might be surprising.
#
# When False, one must call figure() to create new figures. This means that
# gcf() and getfigs() can reference figures created in other cells, and the
# active figure can continue to be edited with pylab/pyplot methods that
# reference the current active figure. This mode facilitates iterative editing
# of figures, and behaves most consistently with other matplotlib backends, but
# figure barriers between cells must be explicit.
# c.InlineBackend.close_figures = True

# Subset of matplotlib rcParams that should be different for the inline backend.
c.InlineBackend.rc = {
    "font.size": 10,
    "savefig.dpi": 72,
    "figure.figsize": (10.0, 6.0),
    "figure.subplot.bottom": 0.125,
}

# ------------------------------------------------------------------------------
# MappingKernelManager configuration
# ------------------------------------------------------------------------------

# A KernelManager that handles notebok mapping and HTTP error handling

# MappingKernelManager will inherit config from: MultiKernelManager

# The max raw message size accepted from the browser over a WebSocket
# connection.
# c.MappingKernelManager.max_msg_size = 65536

# Kernel heartbeat interval in seconds.
# c.MappingKernelManager.time_to_dead = 3.0

# The kernel manager class.  This is configurable to allow subclassing of the
# KernelManager for customized behavior.
# c.MappingKernelManager.kernel_manager_class = 'IPython.zmq.blockingkernelmanager.BlockingKernelManager'

# Delay (in seconds) before sending first heartbeat.
# c.MappingKernelManager.first_beat = 5.0

# ------------------------------------------------------------------------------
# NotebookManager configuration
# ------------------------------------------------------------------------------

# Automatically create a Python script when saving the notebook.
#
# For easier use of import, %run and %load across notebooks, a <notebook-
# name>.py script will be created next to any <notebook-name>.ipynb on each
# save.  This can also be set with the short `--script` flag.
# c.NotebookManager.save_script = False

# The directory to use for notebooks.
# c.NotebookManager.notebook_dir = u''



import unittest

from pydistsim.algorithm import NodeAlgorithm, StatusValues
from pydistsim.algorithm.node_wrapper import NodeAccess
from pydistsim.message import Message
from pydistsim.network import NetworkGenerator
from pydistsim.simulation import Simulation


class TimerAlgorithm(NodeAlgorithm):

    class Status(StatusValues):
        BLOCKER = "BLOCKER"
        OTHER = "OTHER"

    def initializer(self):
        node0 = self.network.nodes_sorted()[0]
        node0.push_to_inbox(Message(meta_header=NodeAlgorithm.INI))
        node0.status = self.Status.BLOCKER

        node1 = self.network.nodes_sorted()[1]
        node1.push_to_inbox(Message(meta_header=NodeAlgorithm.INI))
        node1.status = self.Status.OTHER

    @Status.OTHER
    def spontaneously(self, node: NodeAccess, message: Message):
        n = tuple(node.neighbors())[0]

        self.send(node, 1, n, "ONE")
        self.send(node, 2, n, "TWO")
        self.send(node, 3, n, "THREE")
        self.send(node, 4, n, "FOUR")

    @Status.BLOCKER
    def spontaneously(self, node: NodeAccess, message: Message):
        n = tuple(node.neighbors())[0]

        self.close(node, n)
        self.block_inbox(node, lambda m: m.header != "FOUR")
        self.block_inbox(node, lambda m: m.data != 1)
        node.memory["ALARM"] = False
        self.set_alarm(node, 3)

    @Status.BLOCKER
    def alarm(self, node: NodeAccess, message: Message):
        n = tuple(node.neighbors())[0]
        node.memory["ALARM"] = True
        self.open(node, n)

    @Status.BLOCKER
    def receiving(self, node: NodeAccess, message: Message):
        print(message)

        assert node.memory["ALARM"]
        node.memory[f"RECEIVED_{node.clock}"] = str(message.data)


class TestMsgBlocking(unittest.TestCase):

    def test_blocking(self):
        self.net = NetworkGenerator.generate_complete_network(2)
        sim = Simulation(self.net)
        sim.algorithms = (TimerAlgorithm,)

        sim.run()

        for node in self.net.nodes():
            if node.status == TimerAlgorithm.Status.BLOCKER:
                assert "1" not in node.memory.values()
                assert "4" not in node.memory.values()
                assert "2" in node.memory.values()
                assert "3" in node.memory.values()



"""
Restrictions related to communication among entities/nodes.

These restrictions are related to the communication topology of the underlying graph of the network.
"""

from abc import ABC
from typing import TYPE_CHECKING

from pydistsim.restrictions.base_restriction import CheckableRestriction

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


class CommunicationRestriction(CheckableRestriction, ABC):
    """
    Base class for restrictions related to communication among entities.
    """


class MessageOrdering(CommunicationRestriction):
    """
    In the absence of failure, the messages transmitted by an entity
    to the same out-neighbor will arrive in the same order they are sent.
    """

    help_message = (
        "Message ordering is disabled. Choose a value for `network.behavioral_properties` such that `message_ordering` "
        "is True.\nThe property NetworkBehaviorModel.RandomDelayCommunication should be a good example."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return network.behavioral_properties.message_ordering


class ReciprocalCommunication(CommunicationRestriction):
    """
    For all nodes, the set of out-neighbors is the same as the set of in-neighbors.
    """

    help_message = "The network is not reciprocal. All nodes should have the same set of out-neighbors and "
    "in-neighbors. You may want to use an undirected network or transform the network to be reciprocal (add the "
    "missing edges)."

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        """
        True if the network is undirected or the set of out-neighbors is the same as the set of in-neighbors.
        """
        return not network.is_directed() or all(
            set(network.in_neighbors(node)) == set(network.out_neighbors(node)) for node in network.nodes()
        )


class BidirectionalLinks(ReciprocalCommunication):
    """
    Even if ReciprocalCommunication holds, one node may not know which out-edges correspond
    to which in-edges. ReciprocalCommunication combined with such knowledge is modeled by
    this restriction.
    """

    help_message = (
        "The network can't be directed. Please check the documentation of the network generation method and its "
        "parameters to ensure that the network is undirected."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return not network.is_directed()



from typing import TYPE_CHECKING

from pydistsim.gui.drawing import NODE_COLORS
from pydistsim.utils.tree import get_root_node

if TYPE_CHECKING:
    from pydistsim.network import NetworkType


def show_mst(net: "NetworkType", tree_key="mst"):
    """
    Show tree representation of network.

    :param net: network to show
    :param tree_key: key in nodes memory (dictionary) where parent and
               children data is stored in format:
                {'parent': parent_node,
                 'children': [child_node1, child_node2 ...]}
    """
    nodesToCheck = [(get_root_node(net, tree_key), 0)]
    edgelist = []
    levels = {}  # level of node in tree, root is 0
    while nodesToCheck:
        (node, level) = nodesToCheck.pop()
        edgelist += [(node, child.unbox()) for child in node.memory[tree_key]["children"]]
        levels[node] = NODE_COLORS[level]
        nodesToCheck += [(child.unbox(), level + 1) for child in node.memory[tree_key]["children"]]
    net.show(edge_filter=edgelist, node_colors=levels)
    from matplotlib.pyplot import gca

    gca().set_title("Minimum spanning tree in memory['%s']" % tree_key)



################################################################################
## Form generated from reading UI file 'simulation.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

import icons_rc
from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QAction,
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QCheckBox,
    QDockWidget,
    QFrame,
    QGroupBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMainWindow,
    QMenu,
    QMenuBar,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QToolBar,
    QTreeView,
    QVBoxLayout,
    QWidget,
)


class Ui_SimulationWindow:
    def setupUi(self, SimulationWindow):
        if not SimulationWindow.objectName():
            SimulationWindow.setObjectName("SimulationWindow")
        SimulationWindow.resize(1106, 800)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(100)
        sizePolicy.setHeightForWidth(SimulationWindow.sizePolicy().hasHeightForWidth())
        SimulationWindow.setSizePolicy(sizePolicy)
        SimulationWindow.setMinimumSize(QSize(1096, 800))
        icon = QIcon()
        icon.addFile(":/icons/pydistsim.png", QSize(), QIcon.Normal, QIcon.Off)
        SimulationWindow.setWindowIcon(icon)
        SimulationWindow.setDockOptions(
            QMainWindow.AllowTabbedDocks | QMainWindow.AnimatedDocks | QMainWindow.VerticalTabs
        )
        self.actionRun = QAction(SimulationWindow)
        self.actionRun.setObjectName("actionRun")
        icon1 = QIcon()
        icon1.addFile(":/icons/player_play.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionRun.setIcon(icon1)
        self.actionStep = QAction(SimulationWindow)
        self.actionStep.setObjectName("actionStep")
        icon2 = QIcon()
        icon2.addFile(":/icons/player_fwd.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionStep.setIcon(icon2)
        self.actionReset = QAction(SimulationWindow)
        self.actionReset.setObjectName("actionReset")
        icon3 = QIcon()
        icon3.addFile(":/icons/player_start.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionReset.setIcon(icon3)
        self.actionCopyInspectorData = QAction(SimulationWindow)
        self.actionCopyInspectorData.setObjectName("actionCopyInspectorData")
        self.actionCopyInspectorData.setShortcutContext(Qt.WidgetShortcut)
        self.actionSaveNetwork = QAction(SimulationWindow)
        self.actionSaveNetwork.setObjectName("actionSaveNetwork")
        icon4 = QIcon()
        icon4.addFile(":/icons/filesaveas.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSaveNetwork.setIcon(icon4)
        self.actionOpenNetwork = QAction(SimulationWindow)
        self.actionOpenNetwork.setObjectName("actionOpenNetwork")
        self.actionOpenNetwork.setCheckable(False)
        self.actionOpenNetwork.setChecked(False)
        icon5 = QIcon()
        icon5.addFile(":/icons/fileopen.png", QSize(), QIcon.Normal, QIcon.Off)
        self.actionOpenNetwork.setIcon(icon5)
        self.actionShowLocalizedSubclusters = QAction(SimulationWindow)
        self.actionShowLocalizedSubclusters.setObjectName("actionShowLocalizedSubclusters")
        self.centralwidget = QWidget(SimulationWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.centralwidget.setEnabled(True)
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy1)
        self.horizontalLayout = QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.leftWidget = QWidget(self.centralwidget)
        self.leftWidget.setObjectName("leftWidget")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.leftWidget.sizePolicy().hasHeightForWidth())
        self.leftWidget.setSizePolicy(sizePolicy2)
        self.verticalLayout_3 = QVBoxLayout(self.leftWidget)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.controlGroupBox = QGroupBox(self.leftWidget)
        self.controlGroupBox.setObjectName("controlGroupBox")
        self.controlGroupBox.setMinimumSize(QSize(0, 0))
        self.verticalLayout_5 = QVBoxLayout(self.controlGroupBox)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.widget = QWidget(self.controlGroupBox)
        self.widget.setObjectName("widget")
        self.horizontalLayout_2 = QHBoxLayout(self.widget)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_2 = QLabel(self.widget)
        self.label_2.setObjectName("label_2")

        self.horizontalLayout_2.addWidget(self.label_2)

        self.stepSize = QSpinBox(self.widget)
        self.stepSize.setObjectName("stepSize")
        self.stepSize.setAccelerated(True)
        self.stepSize.setMaximum(999)
        self.stepSize.setValue(1)

        self.horizontalLayout_2.addWidget(self.stepSize)

        self.verticalLayout_5.addWidget(self.widget)

        self.verticalLayout_3.addWidget(self.controlGroupBox)

        self.viewGroupBox = QGroupBox(self.leftWidget)
        self.viewGroupBox.setObjectName("viewGroupBox")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.viewGroupBox.sizePolicy().hasHeightForWidth())
        self.viewGroupBox.setSizePolicy(sizePolicy3)
        self.verticalLayout_2 = QVBoxLayout(self.viewGroupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.networkViewGroup = QGroupBox(self.viewGroupBox)
        self.networkViewGroup.setObjectName("networkViewGroup")
        self.verticalLayout_6 = QVBoxLayout(self.networkViewGroup)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.showNodes = QCheckBox(self.networkViewGroup)
        self.showNodes.setObjectName("showNodes")
        self.showNodes.setChecked(True)

        self.verticalLayout_6.addWidget(self.showNodes)

        self.showEdges = QCheckBox(self.networkViewGroup)
        self.showEdges.setObjectName("showEdges")
        self.showEdges.setChecked(True)

        self.verticalLayout_6.addWidget(self.showEdges)

        self.showMessages = QCheckBox(self.networkViewGroup)
        self.showMessages.setObjectName("showMessages")
        self.showMessages.setChecked(True)

        self.verticalLayout_6.addWidget(self.showMessages)

        self.showLabels = QCheckBox(self.networkViewGroup)
        self.showLabels.setObjectName("showLabels")
        self.showLabels.setChecked(True)

        self.verticalLayout_6.addWidget(self.showLabels)

        self.redrawNetworkButton = QPushButton(self.networkViewGroup)
        self.redrawNetworkButton.setObjectName("redrawNetworkButton")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.redrawNetworkButton.sizePolicy().hasHeightForWidth())
        self.redrawNetworkButton.setSizePolicy(sizePolicy4)

        self.verticalLayout_6.addWidget(self.redrawNetworkButton)

        self.verticalLayout_2.addWidget(self.networkViewGroup)

        self.treeGroupBox = QGroupBox(self.viewGroupBox)
        self.treeGroupBox.setObjectName("treeGroupBox")
        sizePolicy5 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.treeGroupBox.sizePolicy().hasHeightForWidth())
        self.treeGroupBox.setSizePolicy(sizePolicy5)
        self.treeGroupBox.setMinimumSize(QSize(132, 60))
        self.treeGroupBox.setFlat(False)
        self.treeGroupBox.setCheckable(True)
        self.treeGroupBox.setChecked(True)
        self.tree_key = QLineEdit(self.treeGroupBox)
        self.tree_key.setObjectName("tree_key")
        self.tree_key.setGeometry(QRect(42, 20, 71, 20))
        self.tree_key.setFrame(True)
        self.label = QLabel(self.treeGroupBox)
        self.label.setObjectName("label")
        self.label.setGeometry(QRect(10, 22, 31, 16))

        self.verticalLayout_2.addWidget(self.treeGroupBox)

        self.propagationError = QGroupBox(self.viewGroupBox)
        self.propagationError.setObjectName("propagationError")
        self.propagationError.setMinimumSize(QSize(132, 70))
        self.propagationError.setCheckable(True)
        self.propagationError.setChecked(False)
        self.locKey = QLineEdit(self.propagationError)
        self.locKey.setObjectName("locKey")
        self.locKey.setGeometry(QRect(10, 40, 111, 20))
        self.label2 = QLabel(self.propagationError)
        self.label2.setObjectName("label2")
        self.label2.setGeometry(QRect(10, 20, 46, 13))

        self.verticalLayout_2.addWidget(self.propagationError)

        self.verticalLayout_3.addWidget(self.viewGroupBox)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_3.addItem(self.verticalSpacer)

        self.horizontalLayout.addWidget(self.leftWidget)

        self.horizontalSpacer = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.networkDisplayWidget = QWidget(self.centralwidget)
        self.networkDisplayWidget.setObjectName("networkDisplayWidget")
        sizePolicy1.setHeightForWidth(self.networkDisplayWidget.sizePolicy().hasHeightForWidth())
        self.networkDisplayWidget.setSizePolicy(sizePolicy1)
        self.networkDisplayWidget.setMinimumSize(QSize(650, 0))

        self.horizontalLayout.addWidget(self.networkDisplayWidget)

        self.horizontalSpacer_2 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        SimulationWindow.setCentralWidget(self.centralwidget)
        self.toolBar = QToolBar(SimulationWindow)
        self.toolBar.setObjectName("toolBar")
        self.toolBar.setAutoFillBackground(False)
        SimulationWindow.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolBar)
        self.menubar = QMenuBar(SimulationWindow)
        self.menubar.setObjectName("menubar")
        self.menubar.setGeometry(QRect(0, 0, 1106, 21))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuSimulation = QMenu(self.menubar)
        self.menuSimulation.setObjectName("menuSimulation")
        SimulationWindow.setMenuBar(self.menubar)
        self.dockWidget = QDockWidget(SimulationWindow)
        self.dockWidget.setObjectName("dockWidget")
        self.dockWidget.setMinimumSize(QSize(87, 109))
        self.dockWidget.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dockWidget.setAllowedAreas(Qt.RightDockWidgetArea)
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        sizePolicy6 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy6.setHorizontalStretch(0)
        sizePolicy6.setVerticalStretch(0)
        sizePolicy6.setHeightForWidth(self.dockWidgetContents.sizePolicy().hasHeightForWidth())
        self.dockWidgetContents.setSizePolicy(sizePolicy6)
        self.horizontalLayout_3 = QHBoxLayout(self.dockWidgetContents)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.networkInspector = QTreeView(self.dockWidgetContents)
        self.networkInspector.setObjectName("networkInspector")
        self.networkInspector.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.networkInspector.setFrameShape(QFrame.NoFrame)
        self.networkInspector.setProperty("showDropIndicator", False)
        self.networkInspector.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.networkInspector.setAnimated(True)
        self.networkInspector.setWordWrap(True)
        self.networkInspector.setHeaderHidden(True)

        self.horizontalLayout_3.addWidget(self.networkInspector)

        self.dockWidget.setWidget(self.dockWidgetContents)
        SimulationWindow.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.dockWidget)
        self.dockWidget_2 = QDockWidget(SimulationWindow)
        self.dockWidget_2.setObjectName("dockWidget_2")
        self.dockWidget_2.setMinimumSize(QSize(105, 377))
        self.dockWidget_2.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dockWidget_2.setAllowedAreas(Qt.RightDockWidgetArea)
        self.dockWidgetContents_2 = QWidget()
        self.dockWidgetContents_2.setObjectName("dockWidgetContents_2")
        sizePolicy6.setHeightForWidth(self.dockWidgetContents_2.sizePolicy().hasHeightForWidth())
        self.dockWidgetContents_2.setSizePolicy(sizePolicy6)
        self.horizontalLayout_4 = QHBoxLayout(self.dockWidgetContents_2)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.nodeInspector = QTreeView(self.dockWidgetContents_2)
        self.nodeInspector.setObjectName("nodeInspector")
        sizePolicy7 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy7.setHorizontalStretch(0)
        sizePolicy7.setVerticalStretch(0)
        sizePolicy7.setHeightForWidth(self.nodeInspector.sizePolicy().hasHeightForWidth())
        self.nodeInspector.setSizePolicy(sizePolicy7)
        self.nodeInspector.setMinimumSize(QSize(87, 337))
        self.nodeInspector.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.nodeInspector.setFrameShape(QFrame.NoFrame)
        self.nodeInspector.setProperty("showDropIndicator", False)
        self.nodeInspector.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.nodeInspector.setAnimated(True)
        self.nodeInspector.setWordWrap(True)
        self.nodeInspector.setHeaderHidden(True)

        self.horizontalLayout_4.addWidget(self.nodeInspector)

        self.dockWidget_2.setWidget(self.dockWidgetContents_2)
        SimulationWindow.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.dockWidget_2)
        self.dockWidget_3 = QDockWidget(SimulationWindow)
        self.dockWidget_3.setObjectName("dockWidget_3")
        self.dockWidget_3.setMinimumSize(QSize(87, 109))
        self.dockWidget_3.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dockWidget_3.setAllowedAreas(Qt.RightDockWidgetArea)
        self.dockWidgetContents_3 = QWidget()
        self.dockWidgetContents_3.setObjectName("dockWidgetContents_3")
        sizePolicy6.setHeightForWidth(self.dockWidgetContents_3.sizePolicy().hasHeightForWidth())
        self.dockWidgetContents_3.setSizePolicy(sizePolicy6)
        self.horizontalLayout_5 = QHBoxLayout(self.dockWidgetContents_3)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.logListWidget = QListWidget(self.dockWidgetContents_3)
        self.logListWidget.setObjectName("logListWidget")
        self.logListWidget.setFrameShape(QFrame.NoFrame)
        self.logListWidget.setFrameShadow(QFrame.Sunken)
        self.logListWidget.setLineWidth(0)

        self.horizontalLayout_5.addWidget(self.logListWidget)

        self.dockWidget_3.setWidget(self.dockWidgetContents_3)
        SimulationWindow.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.dockWidget_3)
        QWidget.setTabOrder(self.stepSize, self.tree_key)

        self.toolBar.addAction(self.actionOpenNetwork)
        self.toolBar.addAction(self.actionSaveNetwork)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionRun)
        self.toolBar.addAction(self.actionStep)
        self.toolBar.addAction(self.actionReset)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuSimulation.menuAction())
        self.menuFile.addAction(self.actionOpenNetwork)
        self.menuFile.addAction(self.actionSaveNetwork)
        self.menuSimulation.addAction(self.actionRun)
        self.menuSimulation.addAction(self.actionStep)
        self.menuSimulation.addAction(self.actionReset)

        self.retranslateUi(SimulationWindow)

        QMetaObject.connectSlotsByName(SimulationWindow)

    # setupUi

    def retranslateUi(self, SimulationWindow):
        SimulationWindow.setWindowTitle(QCoreApplication.translate("SimulationWindow", "PyDistSim Simulation", None))
        self.actionRun.setText(QCoreApplication.translate("SimulationWindow", "Run", None))
        # if QT_CONFIG(tooltip)
        self.actionRun.setToolTip(QCoreApplication.translate("SimulationWindow", "Run simulation from beginning", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionRun.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+R", None))
        # endif // QT_CONFIG(shortcut)
        self.actionStep.setText(QCoreApplication.translate("SimulationWindow", "Step", None))
        # if QT_CONFIG(tooltip)
        self.actionStep.setToolTip(QCoreApplication.translate("SimulationWindow", "Run next step", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionStep.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+Space", None))
        # endif // QT_CONFIG(shortcut)
        self.actionReset.setText(QCoreApplication.translate("SimulationWindow", "Reset", None))
        # if QT_CONFIG(tooltip)
        self.actionReset.setToolTip(QCoreApplication.translate("SimulationWindow", "Reset simulation", None))
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionReset.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+W", None))
        # endif // QT_CONFIG(shortcut)
        self.actionCopyInspectorData.setText(QCoreApplication.translate("SimulationWindow", "Copy", None))
        # if QT_CONFIG(shortcut)
        self.actionCopyInspectorData.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+C", None))
        # endif // QT_CONFIG(shortcut)
        self.actionSaveNetwork.setText(QCoreApplication.translate("SimulationWindow", "Save", None))
        # if QT_CONFIG(tooltip)
        self.actionSaveNetwork.setToolTip(
            QCoreApplication.translate("SimulationWindow", "Save network in npickle format", None)
        )
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionSaveNetwork.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+S", None))
        # endif // QT_CONFIG(shortcut)
        self.actionOpenNetwork.setText(QCoreApplication.translate("SimulationWindow", "Open", None))
        # if QT_CONFIG(tooltip)
        self.actionOpenNetwork.setToolTip(
            QCoreApplication.translate("SimulationWindow", "Open network from npickle", None)
        )
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionOpenNetwork.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+O", None))
        # endif // QT_CONFIG(shortcut)
        self.actionShowLocalizedSubclusters.setText(
            QCoreApplication.translate("SimulationWindow", "Show localized subclusters", None)
        )
        # if QT_CONFIG(tooltip)
        self.actionShowLocalizedSubclusters.setToolTip(
            QCoreApplication.translate(
                "SimulationWindow",
                "Show localized subclusters based on memory field that has positions and subclusters items.",
                None,
            )
        )
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(shortcut)
        self.actionShowLocalizedSubclusters.setShortcut(QCoreApplication.translate("SimulationWindow", "Ctrl+L", None))
        # endif // QT_CONFIG(shortcut)
        self.controlGroupBox.setTitle(QCoreApplication.translate("SimulationWindow", "Control", None))
        self.label_2.setText(QCoreApplication.translate("SimulationWindow", "Step size:", None))
        self.stepSize.setSpecialValueText(QCoreApplication.translate("SimulationWindow", "All", None))
        self.viewGroupBox.setTitle(QCoreApplication.translate("SimulationWindow", "View", None))
        self.networkViewGroup.setTitle(QCoreApplication.translate("SimulationWindow", "Network", None))
        self.showNodes.setText(QCoreApplication.translate("SimulationWindow", "Nodes", None))
        self.showEdges.setText(QCoreApplication.translate("SimulationWindow", "Edges", None))
        self.showMessages.setText(QCoreApplication.translate("SimulationWindow", "Messages", None))
        self.showLabels.setText(QCoreApplication.translate("SimulationWindow", "Labels", None))
        self.redrawNetworkButton.setText(QCoreApplication.translate("SimulationWindow", "Redraw", None))
        # if QT_CONFIG(tooltip)
        self.treeGroupBox.setToolTip(
            QCoreApplication.translate(
                "SimulationWindow",
                "Enter memory key that has parent and child items.",
                None,
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.treeGroupBox.setTitle(QCoreApplication.translate("SimulationWindow", "Tree", None))
        self.tree_key.setText(QCoreApplication.translate("SimulationWindow", "treeNeighbors", None))
        self.label.setText(QCoreApplication.translate("SimulationWindow", "Key:", None))
        # if QT_CONFIG(tooltip)
        self.propagationError.setToolTip(
            QCoreApplication.translate(
                "SimulationWindow",
                "Enter memory key that has stitch location data.",
                None,
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.propagationError.setTitle(QCoreApplication.translate("SimulationWindow", "Propagation error", None))
        self.locKey.setText(QCoreApplication.translate("SimulationWindow", "convergecastLoc", None))
        self.label2.setText(QCoreApplication.translate("SimulationWindow", "LocKey:", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("SimulationWindow", "toolBar", None))
        self.menuFile.setTitle(QCoreApplication.translate("SimulationWindow", "File", None))
        self.menuSimulation.setTitle(QCoreApplication.translate("SimulationWindow", "Simulation", None))
        self.dockWidget.setWindowTitle(QCoreApplication.translate("SimulationWindow", "Network inspector", None))
        self.dockWidget_2.setWindowTitle(QCoreApplication.translate("SimulationWindow", "Node inspector", None))
        self.dockWidget_3.setWindowTitle(QCoreApplication.translate("SimulationWindow", "Log", None))

    # retranslateUi



from pydistsim.algorithm.node_algorithm import NodeAlgorithm, StatusValues
from pydistsim.algorithm.node_wrapper import NodeAccess
from pydistsim.message import Message
from pydistsim.restrictions.communication import BidirectionalLinks
from pydistsim.restrictions.reliability import TotalReliability
from pydistsim.restrictions.topological import Connectivity, UniqueInitiator


class Flood(NodeAlgorithm):
    default_params = {
        "informationKey": "information",
        "initial_information": "Hello, world!",
    }

    class Status(StatusValues):
        INITIATOR = "INITIATOR"
        IDLE = "IDLE"
        DONE = "DONE"

    S_init = (Status.INITIATOR, Status.IDLE)
    S_term = (Status.DONE,)

    algorithm_restrictions = (
        BidirectionalLinks,
        TotalReliability,
        Connectivity,
        UniqueInitiator,
    )

    def initializer(self):
        for node in self.network.nodes():
            node.status = self.Status.IDLE

        ini_node = self.network.nodes_sorted()[0]
        ini_node.push_to_inbox(Message(meta_header=NodeAlgorithm.INI, destination=ini_node))
        ini_node.status = self.Status.INITIATOR
        ini_node.memory[self.informationKey] = self.initial_information

    @Status.INITIATOR
    def spontaneously(self, node: NodeAccess, message: Message):
        self.send(
            node,
            data=node.memory[self.informationKey],
            destination=list(node.neighbors()),
            header="Information",
        )
        node.status = self.Status.DONE

    @Status.IDLE
    def receiving(self, node: NodeAccess, message: Message):
        if message.header == "Information":
            node.memory[self.informationKey] = message.data
            destination_nodes = node.neighbors()
            # send to every neighbor, except the original sender
            destination_nodes.remove(message.source)
            if destination_nodes:
                self.send(
                    node,
                    destination=destination_nodes,
                    header="Information",
                    data=message.data,
                )
        node.status = self.Status.DONE

    @Status.DONE
    def default(self, *args, **kwargs):
        "Do nothing, for all inputs."
        pass



from abc import ABC, abstractmethod
from math import inf as Inf

import png
from numpy import ones, sign, sqrt, uint8, vstack
from numpy.random import rand

from pydistsim.logging import logger


class Environment(ABC):
    """
    Environment abstract base class.

    This class represents an abstract base class for environments in the PyDistSim framework.
    """

    @abstractmethod
    def is_space(self, xy):
        """Check if the given coordinates represent a valid space in the environment.

        This method should be implemented by subclasses to determine whether the given coordinates
        represent a valid space in the environment.

        :param xy: The coordinates to check.
        :return: True if the coordinates represent a valid space, False otherwise.
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def are_visible(self, xy1, xy2):
        """Check if two coordinates are visible to each other in the environment.

        This method should be implemented by subclasses to determine whether the two given coordinates
        are visible to each other in the environment.

        :param xy1: The first set of coordinates.
        :param xy2: The second set of coordinates.
        :return: True if the coordinates are visible to each other, False otherwise.
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def find_random_pos(self, n=100):
        """
        Returns a random position in the environment.

        :param n: The maximum number of iterations to find a free space.
        :type n: int
        :return: The random position found.
        :rtype: tuple
        """
        raise NotImplementedError


class Environment2D(Environment):
    """
    Base class for 2D environment.
    The Environment2D allows to define map and scale of 2D environment.

    :param path: Optional. The path to an image file to load as the environment. If not provided, a default environment will be created.
    :param scale: Optional. The scale factor for the environment image. If not provided, the default scale factor will be used.
    :param shape: Optional. The shape of the environment. If not provided, the default shape will be used.
    """

    def __init__(self, path="", scale=None, shape=None):
        shape = shape if shape else (600, 600)
        if path:
            try:
                r = png.Reader(path)
                planes = r.read()[3]["planes"]
                self._image = vstack(map(uint8, r.asDirect()[2]))[:, ::planes]
                self._image = self._image[::-1, :]  # flip-up-down
                assert (r.height, r.width) == self._image.shape
            except OSError:
                logger.exception("Can't open {} creating new default environment.", path)

                self._image = uint8(ones(shape) * 255)
        else:
            self._image = uint8(ones(shape) * 255)

        self._dim = 2
        scale = not scale and 1 or int(scale)
        if scale > 1:
            raise NotImplementedError

    @property
    def image(self):
        return self._image

    @property
    def dim(self):
        return self._dim

    @image.setter
    def image(self, image):
        # Immutable object
        raise AttributeError("Can't set attribute 'image'")

    @dim.setter
    def dim(self, dim):
        # Immutable object
        raise AttributeError("Can't set attribute 'dim'")

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def is_space(self, xy):
        """
        Returns true if selected space (x,y) is space. If point xy
        is exactly on the edge or crossing check surrounding pixels.

        :param xy: A tuple representing the coordinates (x, y) of the space to check.
        :type xy: tuple
        :return: True if the selected space is a space, False otherwise.
        :rtype: bool
        """
        x, y = xy
        h, w = self._image.shape
        if x < 0 or x > w or y < 0 or y > h:
            return False
        check = True
        points = [xy]
        if xy[0] % 1 == 0:
            points.append([xy[0] - 1, xy[1]])
        if xy[1] % 1 == 0:
            points.append([xy[0], xy[1] - 1])
        if xy[0] % 1 == 0 and xy[1] % 1 == 0:
            points.append([xy[0] - 1, xy[1] - 1])
        try:
            for p in points:
                check = check and self._image[int(p[1]), int(p[0])] != 0
        except IndexError:
            check = False
        return check

    def are_visible(self, xy0, xy1):
        """
        Returns true if there is line of sight between source (x0,y0) and
        destination (x1,y1).

        :param xy0: Tuple representing the coordinates of the source point (x0, y0).
        :type xy0: tuple
        :param xy1: Tuple representing the coordinates of the destination point (x1, y1).
        :type xy1: tuple
        :return: True if there is line of sight between the source and destination points, False otherwise.
        :rtype: bool

        This is a floating point version of the Bresenham algorithm that does not spread on diagonal pixels.
        """
        x = x0 = xy0[0]
        y = y0 = xy0[1]
        x1 = xy1[0]
        y1 = xy1[1]
        d = sqrt(pow(x1 - x0, 2) + pow(y1 - y0, 2))
        incrE = (x1 - x0) / d  # incrE is cos in direction of x axis
        incrN = (y1 - y0) / d  # incrN is sin in direction of y axis

        # check if pixel (x,y) is target pixel (x1,y1) or
        # if float (x,y) is on N or E edge then check also W or S neighbor
        while not (
            (int(x) == int(x1) or (x % 1 == 0 and int(x) - 1 == int(x1)))
            and (int(y) == int(y1) or (y % 1 == 0 and int(y) - 1 == int(y1)))
        ):
            if incrE > 0:
                dx = 1 - x % 1
            else:
                dx = x % 1 == 0 and 1.0 or x % 1
            if incrN > 0:
                dy = 1 - y % 1
            else:
                dy = y % 1 == 0 and 1.0 or y % 1

            # check whether the path will hit first E/W or N/S pixel edge
            # by calculating length of paths
            if incrE != 0:
                cx = abs(dx / incrE)
            else:
                cx = Inf
            if incrN != 0:
                cy = abs(dy / incrN)
            else:
                cy = Inf

            # if path needed to hit N/S edge is longer than E/W
            if cx < cy:
                x = round(x + sign(incrE) * dx)  # spread on E
                y = y + cx * incrN
            else:
                x = x + cy * incrE
                y = round(y + sign(incrN) * dy)  # spread on N

            # logger.debug('x = {}, y = {}', str(x), str(y))
            if not self.is_space([x, y]):
                return False
        return True

    def find_random_pos(self, n=100):
        """
        Returns a random position in the environment.

        :param n: The maximum number of iterations to find a free space.
        :type n: int
        :return: The random position found.
        :rtype: tuple
        """
        n_init = n
        while n > 0:
            pos = rand(self._dim) * tuple(reversed(self._image.shape))
            if self.is_space(pos):
                break
            n -= 1
        logger.trace("Random position found in {} iterations.", (n_init - n))
        return pos



from contextlib import contextmanager

from pydistsim.demo_algorithms.broadcast import Flood
from pydistsim.network import NetworkGenerator
from pydistsim.observers import (
    AlgorithmObserver,
    NetworkObserver,
    NodeObserver,
    ObservableEvents,
    SimulationObserver,
)
from pydistsim.simulation import Simulation
from pydistsim.utils.testing import PyDistSimTestCase


class ObserverForTests(AlgorithmObserver, SimulationObserver, NetworkObserver, NodeObserver):
    """
    A test observer that raises an exception when notified.
    """

    events = list(ObservableEvents._all)

    class ObserverNotified(Exception):
        def __init__(self, event):
            self.event = event

    def __init__(self) -> None:
        self.raise_e = None

    @contextmanager
    def do_raise(self, event):
        try:
            self.raise_e = event
            yield self
        finally:
            self.raise_e = None

    def on_added(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.added:
            raise self.ObserverNotified(ObservableEvents.added)

    def on_step_done(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.step_done:
            raise self.ObserverNotified(ObservableEvents.step_done)

    def on_message_sent(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.message_sent:
            raise self.ObserverNotified(ObservableEvents.message_sent)

    def on_sim_state_changed(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.sim_state_changed:
            raise self.ObserverNotified(ObservableEvents.sim_state_changed)

    def on_algorithm_started(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.algorithm_started:
            raise self.ObserverNotified(ObservableEvents.algorithm_started)

    def on_algorithm_finished(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.algorithm_finished:
            raise self.ObserverNotified(ObservableEvents.algorithm_finished)

    def on_network_changed(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.network_changed:
            raise self.ObserverNotified(ObservableEvents.network_changed)

    def on_message_delivered(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.message_delivered:
            raise self.ObserverNotified(ObservableEvents.message_delivered)

    def on_node_status_changed(self, *args, **kwargs):
        if self.raise_e == ObservableEvents.node_status_changed:
            raise self.ObserverNotified(ObservableEvents.node_status_changed)


class TestObserver(PyDistSimTestCase):
    def setUp(self):
        super().setUpClass()
        self.observer = ObserverForTests()
        self.net = NetworkGenerator(10, directed=False).generate_random_network()

        self.sim = Simulation(self.net)
        self.sim.algorithms = ((Flood, {"informationKey": "greet"}),)

        self.initiator = self.net.nodes_sorted()[0]
        self.initiator.memory["greet"] = "HELLO"

    def test_added(self):
        with self.observer.do_raise(ObservableEvents.added):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.add_observers(self.observer)

        self.assertEqual(context.exception.event, ObservableEvents.added)

    def test_step_done(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.step_done):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.run(1)

        self.assertEqual(context.exception.event, ObservableEvents.step_done)

    def test_algorithm_started(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.algorithm_started):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.run(1)

        self.assertEqual(context.exception.event, ObservableEvents.algorithm_started)

    def test_algorithm_finished(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.algorithm_finished):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.run(100_000)

        self.assertEqual(context.exception.event, ObservableEvents.algorithm_finished)

    def test_network_changed(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.network_changed):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.network = self.net

        self.assertEqual(context.exception.event, ObservableEvents.network_changed)

    def test_sim_state_changed(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.sim_state_changed):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.run(100_000)

        self.assertEqual(context.exception.event, ObservableEvents.sim_state_changed)

    def test_message_sent(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.message_sent):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.run(3)

        self.assertEqual(context.exception.event, ObservableEvents.message_sent)

    def test_message_delivered(self):
        self.sim.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.message_delivered):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                self.sim.run(1)

        self.assertEqual(context.exception.event, ObservableEvents.message_delivered)

    def test_node_status_changed(self):
        some_node = self.net.nodes_sorted()[0]
        some_node.add_observers(self.observer)
        with self.observer.do_raise(ObservableEvents.node_status_changed):
            with self.assertRaises(ObserverForTests.ObserverNotified) as context:
                some_node.status = "NEW"

        self.assertEqual(context.exception.event, ObservableEvents.node_status_changed)


class TestWrongEvent(PyDistSimTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.observer = ObserverForTests()
        cls.net = NetworkGenerator(10, directed=False).generate_random_network()

    def test_wrong_event(self):
        # Nothing should happen
        self.net.add_observers(self.observer)
        self.net.notify_observers("test", None)



class StitchSubclusterSelectorBase:
    """Subclasses of this class are used to make choice which subclusters
    will be stitched next based on criteria."""

    def __init__(self, *args, **kwargs):
        pass

    def select(self, dst, src, stitched, is_intra=False):
        raise NotImplementedError


class MaxCommonNodeSelector(StitchSubclusterSelectorBase):
    """
    Return subclusters that have maximum common node count.
    """

    def __init__(self, cn_count_treshold=0):
        self.cn_count_treshold = cn_count_treshold

    def select(self, dst, src, stitched, is_intra=False):
        cn_count_max = 0
        dstSubIndex = None
        srcSubIndex = None
        for src_index, src_sc in enumerate(src):
            for dst_index, dst_sc in enumerate(dst):
                # same subcluster
                if is_intra and src_index == dst_index:
                    continue
                # already stitched
                if (dst_index, src_index) in stitched:
                    continue
                # src_sc is subset of dst_sc mark src_sc stitched with dst
                if set(src_sc.keys()) <= set(dst_sc.keys()):
                    stitched[(dst_index, src_index)] = ()
                    continue

                commonNodes = [node for node in list(dst_sc.keys()) if node in list(src_sc.keys())]
                cn_count = len(commonNodes)
                if cn_count > cn_count_max and cn_count >= self.cn_count_treshold:
                    cn_count_max = cn_count
                    dstSubIndex = dst_index
                    srcSubIndex = src_index

        # in intraStitch make sure that dstSub is larger one
        if is_intra and cn_count_max != 0 and len(src[srcSubIndex]) > len(dst[dstSubIndex]):
            return (srcSubIndex, dstSubIndex)  # swapped src and dst indexes
        return (dstSubIndex, srcSubIndex)



"""
This module contains the :class:`Message` class, which is used to represent messages in the simulation.
"""

from copy import copy, deepcopy
from enum import StrEnum
from typing import Generic, TypeVar


class MetaHeader(StrEnum):
    NORMAL_MESSAGE = "NORMAL_MESSAGE"
    INITIALIZATION_MESSAGE = "INITIALIZATION_MESSAGE"
    ALARM_MESSAGE = "ALARM_MESSAGE"


T = TypeVar("T")


class Message(Generic[T]):
    """
    The Message class is used to represent messages in the simulation.

    :param destination: The destination of the message.
    :type destination: T
    :param header: The header of the message.
    :type header: str
    :param data: The data associated with the message.
    :type data: dict
    :param source: The source of the message.
    :type source: T
    :param meta_header: The meta header of the message.
    :type meta_header: MetaHeader
    :param meta_data: The meta data associated with the message. This is meant to be used by the simulation.
    :type meta_data: dict
    """

    __slots__ = [
        "destination",
        "header",
        "data",
        "source",
        "meta_header",
        "meta_data",
        "_internal_id",
    ]

    META_HEADERS = MetaHeader
    next_message_id = 1

    def __init__(
        self,
        destination: T = None,
        header=None,
        data=None,
        **kwargs,
    ):
        self.destination: T = destination
        self.header = header or "NO HEADER"
        self.data = data

        self.source: T = kwargs.get("source", None)
        self.meta_header = kwargs.get("meta_header", MetaHeader.NORMAL_MESSAGE)
        self.meta_data = kwargs.get("meta_data", {})

        self._internal_id = self.__class__.next_message_id
        self.__class__.next_message_id += 1

    def __repr__(self):
        destination = self.destination
        if self.destination is None:
            destination = "Broadcasted"
        elif isinstance(self.destination, list) and len(self.destination) == 1 and self.destination[0] is None:
            destination = "Broadcasting"
        return (
            "\n------ Message '%s' ------ \n     source = %s \ndestination = %s"
            " \n     header = '%s' \nid(message) = 0x%x>"
        ) % (self.meta_header, self.source, destination, self.header, id(self))

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]

        # Shallow copy of the object
        copy_m = copy(self)
        memo[id(self)] = copy_m

        copy_m._internal_id = self.__class__.next_message_id
        self.__class__.next_message_id += 1

        # Deep copy of the mutable attributes
        copy_m.source = deepcopy(self.source, memo)
        copy_m.destination = deepcopy(self.destination, memo)
        copy_m.data = deepcopy(self.data, memo)
        copy_m.meta_data = deepcopy(self.meta_data, memo)

        return copy_m

    def copy(self):
        """
        Create a copy of the Message object.
        """
        # nodes are protected from copying by __deepcopy__()
        copy_data = deepcopy(self.data)
        new_message = copy(self)
        new_message.data = copy_data
        new_message._internal_id = self.__class__.next_message_id
        self.__class__.next_message_id += 1

        return new_message



"""
Custom exceptions for the PyDistSim package.
"""

from enum import StrEnum


class MessageUndeliverableException(Exception):
    def __init__(self, e, message):
        self.e = e
        self.message = message

    def __str__(self):
        return self.e + repr(self.message)


class PyDistSimException(Exception):
    ERRORS: type[StrEnum]

    def __init__(self, type_):
        if isinstance(type_, self.ERRORS):
            self.message = type_.value
        else:
            self.message = "Unknown error."

    def __str__(self):
        return self.message


class NetworkException(PyDistSimException):
    """
    Exception class for network-related errors.
    """

    class ERRORS(StrEnum):
        NODE = "Node is already in another network."
        NODE_SPACE = "Given position is not free space."
        NODE_NOT_IN_NET = "Node not in network."


class SimulationException(PyDistSimException):
    """
    Exception class for simulation-related errors.
    """

    class ERRORS(StrEnum):
        ALGORITHM = (
            "Algorithms must be in tuple (AlgorithmClass,)"
            " or in form: ((AlgorithmClass, params_dict),)."
            "AlgorithmClass should be subclass of BaseAlgorithm"
        )
        ALGORITHM_NOT_FOUND = "Algorithm not found in simulation. Please make sure it is added to the simulation."



from pydistsim.algorithm import NetworkAlgorithm


class ReadSensors(NetworkAlgorithm):
    """Read all sensors and save data in memory for latter usage."""

    required_params = ()
    default_params = {"sensorReadingsKey": "sensorReadings"}

    def run(self):
        for node in self.network.nodes():
            node.memory.update({self.sensorReadingsKey: node.compositeSensor.read()})



from collections.abc import Callable
from copy import copy, deepcopy
from typing import TYPE_CHECKING, Optional, TypeVar

from pydistsim.logging import LogLevels, logger
from pydistsim.network.sensor import CompositeSensor
from pydistsim.observers import ObservableEvents, ObserverManagerMixin

if TYPE_CHECKING:
    from pydistsim.message import Message
    from pydistsim.network.network import NetworkType
    from pydistsim.network.sensor import Sensor


T = TypeVar("T")


class BlockableQueue(list[T]):

    def __init__(self):
        super().__init__()
        self._filters: list[Callable[[T], bool]] = []

    def add_block(self, filter_func: Callable[[T], bool]) -> Callable[[T], bool]:
        self._filters.append(filter_func)
        return filter_func

    def remove_filter(self, filter_func: Callable[[T], bool]):
        self._filters.remove(filter_func)

    def pop(self):
        for i in reversed(range(super().__len__())):
            if all(f(self[i]) for f in self._filters):
                return super().pop(i)
        raise IndexError("pop from empty or blocked list")

    def __iter__(self):
        return filter(lambda x: all(f(x) for f in self._filters), super().__iter__())

    def __bool__(self):
        return self.__len__() > 0

    def __len__(self):
        return len(tuple(iter(self)))


class Node(ObserverManagerMixin):
    """
    Represents a node in a network.

    :param network: Optional network object to which the node belongs.
    :type network: Network, optional
    :param commRange: Communication range of the node.
    :type commRange: int, optional
    :param sensors: Tuple of sensor types or names.
    :type sensors: tuple[type[Sensor] | str], optional
    :param kwargs: Additional keyword arguments.
    """

    next_node_id = 1

    def __init__(
        self,
        network: Optional["NetworkType"] = None,
        commRange: None | int = None,
        sensors: None | tuple[type["Sensor"] | str] = None,
        **kwargs,
    ):
        super().__init__()
        self._compositeSensor = CompositeSensor(self, sensors or ())
        "Object that collects the data from all available sensors."

        self.network = network
        "The network to which the node belongs."

        self._internal_id = self.__class__.next_node_id
        "Internal ID of the node. Only used for internal simulation purposes."
        self.__class__.next_node_id += 1

        self._commRange = commRange or 100
        self._inboxDelay = True
        self._status = None
        self.outbox: BlockableQueue["Message"] = BlockableQueue()
        self._inbox: BlockableQueue["Message"] = BlockableQueue()
        self.memory = {}
        self.clock = 0

    def __repr__(self):
        return self.__repr_str__(self._internal_id)

    @staticmethod
    def __repr_str__(id):
        return "<Node id=%s>" % id

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]

        # Shallow copy of the object
        copy_n = copy(self)
        memo[id(self)] = copy_n

        copy_n._internal_id = self.__class__.next_node_id
        self.__class__.next_node_id += 1

        # Deep copy of the mutable attributes
        copy_n._compositeSensor = deepcopy(self._compositeSensor, memo)
        copy_n.network = deepcopy(self.network, memo)
        copy_n.outbox = deepcopy(self.outbox, memo)
        copy_n._inbox = deepcopy(self._inbox, memo)
        copy_n.memory = deepcopy(self.memory, memo)

        copy_n.clear_observers()

        return copy_n

    @property
    def status(self):
        """
        Get the status of the node.

        :return: The status of the node.
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status: str):
        """
        Set the status of the node.

        :param status: The status to be set.
        :type status: str
        """
        self.notify_observers(
            ObservableEvents.node_status_changed,
            node=self,
            previous_status=self._status,
            new_status=status,
        )
        self._status = status

    def reset(self):
        """
        Reset the node's state.

        Clears the outbox, inbox, status, and memory of the node.
        """
        self.outbox: BlockableQueue["Message"] = BlockableQueue()
        self._inbox: BlockableQueue["Message"] = BlockableQueue()
        self._status = None
        self.memory = {}
        self.clock = 0

    def block_inbox(self, filter_func: Callable[["Message"], bool]) -> Callable[["Message"], bool]:
        """
        Block messages from being received by the node.

        :param filter_func: The filter function to be used.
        :type filter_func: function
        :return: The filter function.
        """
        return self._inbox.add_block(filter_func)

    def unblock_inbox(self, filter_func: Callable[["Message"], bool]):
        """
        Unblock messages from being received by the node.

        :param filter_func: The filter function to be removed.
        :type filter_func: function
        """
        self._inbox.remove_filter(filter_func)

    def receive(self):
        """
        Pop message from inbox but only if it has been there at least one step.

        Messages should be delayed for one step for visualization purposes.
        Messages are processed without delay only if they are pushed into empty
        inbox. So if inbox is empty when push_to_inbox is called _inboxDelay is
        set to True.

        This method is used only internally and is not supposed to be used
        inside algorithms.

        :return: The received message, or None if no message is available.
        :rtype: Message or None
        """
        if self._inbox and not self._inboxDelay:
            # TODO: implement precedence in message type: Spontaneously > Alarm > Receiving
            message = self._inbox.pop()
            logger.debug("Node {} received message {}", self._internal_id, message.__repr__())
        else:
            message = None
        self._inboxDelay = False
        return message

    @property
    def inbox(self):
        """
        Get the inbox of the node.

        :return: The inbox of the node.
        :rtype: list[Message]
        """
        return self._inbox

    def push_to_inbox(self, message: "Message[Node]"):
        """
        Push a message to the inbox of the node.

        :param message: The message to be pushed to the inbox.
        :type message: Message
        """
        self._inboxDelay = self._inboxDelay or not self._inbox
        self._inbox.insert(0, message)
        logger.debug("Message delivered to {}", self)
        self.notify_observers(ObservableEvents.message_delivered, message)

    def push_to_outbox(self, message: "Message[Node]", destination: "Node"):
        """
        Push a message to the outbox of the node.

        :param message: The message to be pushed to the outbox.
        :type message: Message
        :param destination: The destination node of the message.
        :type destination: Node
        """
        message.destination = destination
        self.outbox.insert(0, message)
        logger.debug("Node {} sent message {}.", self._internal_id, message.__repr__())
        self.notify_observers(ObservableEvents.message_sent, message)

    @property
    def compositeSensor(self):
        """
        Get the composite sensor of the node.

        :return: The composite sensor of the node.
        :rtype: CompositeSensor
        """
        return self._compositeSensor

    @compositeSensor.setter
    def compositeSensor(self, compositeSensor):
        """
        Set the composite sensor of the node.

        :param compositeSensor: The composite sensor to be set.
        :type compositeSensor: CompositeSensor
        """
        self._compositeSensor = CompositeSensor(self, compositeSensor)

    @property
    def sensors(self):
        """
        Get the sensors of the node.

        :return: The sensors of the node.
        :rtype: list[Sensor]
        """
        return self._compositeSensor.sensors

    @sensors.setter
    def sensors(self, sensors: tuple[type["Sensor"] | str]):
        """
        Set the sensors of the node.

        :param sensors: The sensors to be set.
        :type sensors: tuple[type[Sensor] | str]
        """
        self._compositeSensor = CompositeSensor(self, sensors)

    @property
    def commRange(self) -> int:
        """
        Get the communication range of the node.

        :return: The communication range of the node.
        :rtype: int
        """
        return self._commRange

    @commRange.setter
    def commRange(self, commRange: int):
        """
        Set the communication range of the node.

        :param commRange: The communication range to be set.
        :type commRange: int
        """
        self._commRange = commRange
        if self.network and hasattr(self.network, "recalculate_edges"):
            self.network.recalculate_edges([self])

    def get_log(self):
        """
        Get the log messages stored in the node's memory.

        :return: The log messages.
        :rtype: list[tuple[LogLevels, str, dict]]
        """
        if "log" not in self.memory:
            self.memory["log"] = []
        return self.memory["log"]

    def log(self, message: str, level: LogLevels = LogLevels.WARNING):
        """
        Insert a log message in the node's memory.

        :param message: The log message to be inserted.
        :type message: str
        :param level: The log level of the message.
        :type level: LogLevels, optional
        """
        assert isinstance(message, str)
        context = {
            "algorithm": str(self.network.simulation.get_current_algorithm()),
            "algorithmState": self.network.simulation.algorithmState,
            "clock": self.clock,
        }
        if "log" not in self.memory:
            self.memory["log"] = [(level, message, context)]
        else:
            self.memory["log"].append((level, message, context))

    def get_dic(self):
        """
        Get the node's information as a dictionary.

        :return: The node's information.
        :rtype: dict
        """
        return {
            "1. info": {
                "id": self._internal_id,
                "status": self.status or "",
                "position": self.network.pos[self],
                "orientation": self.network.ori[self],
            },
            "2. communication": {
                "range": self.commRange,
                "inbox": self.box_as_dic("inbox"),
                "outbox": self.box_as_dic("outbox"),
            },
            "3. memory": self.memory,
            "4. sensors": {
                sensor.name(): (
                    "%s(%.3f)"
                    % (
                        sensor.probabilityFunction.name,
                        sensor.probabilityFunction.scale,
                    )
                    if hasattr(sensor, "probabilityFunction") and sensor.probabilityFunction is not None
                    else ("", 0)
                )
                for sensor in self.compositeSensor.sensors
            },
            "5. clock": self.clock,
        }

    def box_as_dic(self, box: str):
        """
        Convert a message box to a dictionary representation.

        :param box: The message box to be converted.
        :type box: str
        :return: The dictionary representation of the message box.
        :rtype: dict
        """
        messagebox = self.__getattribute__(box)
        dic = {}
        for i, message in enumerate(messagebox):
            dic.update(
                {
                    "%d. Message"
                    % (i + 1,): {
                        "1 header": message.header,
                        "2 source": message.source,
                        "3 destination": message.destination,
                        "4 data": message.data,
                    }
                }
            )
        return dic

    def unbox(self):
        return self



import os
import unittest

from pydistsim.demo_algorithms.broadcast import Flood
from pydistsim.network import Network, NetworkGenerator, NetworkType
from pydistsim.utils.npickle import read_npickle, write_npickle


class TestPickle(unittest.TestCase):
    def setUp(self):
        self.nets = []
        for size in (
            10,
            100,
            500,
        ):  # TODO at npickle.py says a recursion error occurs with 6000, couldnt reproduce it
            net = NetworkGenerator(size).generate_random_network()

            net.elected = net.nodes_sorted()[0]
            net.elected.memory[f"some_key_{size}"] = f"some_value_{size}"

            self.nets.append((size, net))

    def test_write_read(self):
        for i, (size, net) in enumerate(self.nets):
            net: Network
            with self.subTest(net=net):
                try:
                    write_npickle(net, f"net_test_{i}.tar.gz")

                    assert os.path.isfile(f"net_test_{i}.tar.gz"), "The file has not been created"

                    net_from_file: "NetworkType" = read_npickle(f"net_test_{i}.tar.gz")

                    assert type(net) == type(net_from_file), "The type of the object is not the same."

                    assert len(net.nodes()) == len(net_from_file.nodes()), "The number of nodes is not the same."

                    assert (
                        net.elected._internal_id == net_from_file.elected._internal_id
                    ), "The elected node is not the same."
                    assert (
                        net.elected.memory[f"some_key_{size}"] == net_from_file.elected.memory[f"some_key_{size}"]
                    ), "The value of the elected node is not the same."

                finally:
                    os.remove(f"net_test_{i}.tar.gz")
                    assert not os.path.isfile(f"net_test_{i}.tar.gz"), "The file has not been deleted"

    def test_read_not_found(self):
        with self.assertRaises(OSError):
            read_npickle("file_not_found.tar.gz")



"""
Simulation module for the PyDistSim package.
"""

import inspect
from copy import copy, deepcopy
from typing import TYPE_CHECKING, Optional

from pydistsim._exceptions import SimulationException
from pydistsim.algorithm import BaseAlgorithm
from pydistsim.logging import logger
from pydistsim.observers import (
    NetworkObserver,
    ObservableEvents,
    ObserverManagerMixin,
    SimulationObserver,
)

if TYPE_CHECKING:
    from pydistsim.network import NetworkType


AlgorithmsParam = tuple[type["BaseAlgorithm"] | tuple[type["BaseAlgorithm"], dict]]


class Simulation(ObserverManagerMixin):
    """
    Controls single network algorithm and node algorithms simulation.
    It is responsible for visualization and logging, also.

    :param network: The network object representing the simulation network.
    :type network: NetworkType
    :param algorithms: The algorithms to be executed on the network.
    :type algorithms: AlgorithmsParam, optional
    :param check_restrictions: Whether to check restrictions during the simulation.
    :type check_restrictions: bool, optional
    :param kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        network: "NetworkType",
        algorithms: AlgorithmsParam | None = None,
        check_restrictions: bool = True,
        **kwargs,
    ):
        super().__init__()

        self._network = network
        self._network.simulation = self
        self._algorithms = ()
        if algorithms is not None:
            self.algorithms = algorithms

        self.algorithmState = {"index": 0, "step": 1, "finished": False}
        self.stepsLeft = 0
        self.check_restrictions = check_restrictions

        logger.info("Simulation {} created successfully.", hex(id(self)))

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]

        # Shallow copy of the object
        copy_s = type(self)(deepcopy(self._network, memo))
        memo[id(self)] = copy_s

        # Shallow copy of the immutable attributes
        copy_s.algorithms = self._algorithms_param
        copy_s.algorithmState = copy(self.algorithmState)
        copy_s.stepsLeft = self.stepsLeft
        copy_s.check_restrictions = self.check_restrictions

        # Deep copy of the mutable attributes
        copy_s.algorithmState = deepcopy(self.algorithmState, memo)
        copy_s.stepsLeft = self.stepsLeft

        copy_s.clear_observers()

        return copy_s

    def run(self, steps=0):
        """
        Run simulation from the current state.

        :param steps: Number of steps to run the simulation.
                      If steps = 0, it runs until all algorithms are finished.
                      If steps > 0, the simulation is in stepping mode.
                      If steps > number of steps to finish the current algorithm, it finishes it.
        :type steps: int
        """
        self.stepsLeft = steps
        for _ in range(len(self.algorithms) * len(self.network)):
            algorithm: Optional["BaseAlgorithm"] = self.get_current_algorithm()
            if not algorithm:
                logger.info(
                    "Simulation has finished. There are no algorithms left to run. "
                    "To run it from the start use sim.reset()."
                )
                self.notify_observers(ObservableEvents.sim_state_changed, self)
                break
            algorithm.add_observers(*self.observers)
            self._run_algorithm(algorithm)
            self.notify_observers(ObservableEvents.sim_state_changed, self)
            if self.stepsLeft <= 0:
                break

    def run_step(self):
        """
        Run a single step of the simulation.

        This is equivalent to calling sim.run(1).
        """
        self.run(1)

    def _run_algorithm(self, algorithm: BaseAlgorithm):
        """
        Run the given algorithm on the given network.

        Update stepsLeft and sim.algorithmState['step'].
        If stepsLeft hit 0 it may return unfinished.

        :param algorithm: The algorithm to run on the network.
        """
        for _ in range(1000 * len(self.network)):
            logger.debug(
                "[{}] Step {} started",
                algorithm.name,
                self.algorithmState["step"],
            )
            algorithm.step(self.check_restrictions, self.algorithmState["step"])
            self.stepsLeft -= 1
            self.algorithmState["step"] += 1
            self.network.increment_node_clocks()
            logger.debug(
                "[{}] Step {} finished",
                algorithm.name,
                self.algorithmState["step"],
            )

            if algorithm.is_halted():
                break  # algorithm finished
            if self.stepsLeft == 0:
                return  # stepped execution finished

        self.notify_observers(ObservableEvents.algorithm_finished, algorithm)
        logger.info("[{}] Algorithm finished", algorithm.name)
        self.algorithmState["finished"] = True

    def reset(self):
        """
        Reset the simulation.
        """
        logger.debug("Resetting simulation.")
        self.algorithmState = {"index": 0, "step": 1, "finished": False}
        self._network.reset()
        if self._algorithms:
            for algorithm in self._algorithms:
                algorithm.reset()

    def is_halted(self):
        """
        Check if simulation has come to an end or deadlock,
        i.e. there are no messages to pass and no alarms set.

        A not-started algorithm is considered halted. If there are
        no algorithms left to run, the simulation is also considered halted.

        :return: True if the algorithm is halted, False otherwise.
        :rtype: bool
        """
        algorithm: Optional["BaseAlgorithm"] = self.get_current_algorithm()
        return algorithm is None or algorithm.is_halted()

    @property
    def network(self):
        """
        Get the network associated with the simulation.
        """
        return self._network

    @network.setter
    def network(self, network: "NetworkType"):
        """
        Set the network for the simulation.

        :param network: The network object to set.
        :type network: NetworkType
        """
        self._network.simulation = None  # remove reference to this simulation in the old network
        self._network.clear_observers()

        self._network = network
        self._network.simulation = self
        self.notify_observers(ObservableEvents.network_changed, self)
        self._copy_observers_to_network()

    def add_observers(self, *observers: "SimulationObserver"):
        super().add_observers(*observers)
        self._copy_observers_to_network()

    def _copy_observers_to_network(self):
        self.network.add_observers(*(observer for observer in self.observers if isinstance(observer, NetworkObserver)))

    #### Algorithm relation methods ####

    def get_current_algorithm(self) -> BaseAlgorithm | None:
        """
        Try to return the current algorithm based on the algorithmState.

        :return: The current algorithm.
        :rtype: BaseAlgorithm or None

        :raises NetworkException: If there are no algorithms defined in the network.
        """
        if len(self.algorithms) == 0:
            logger.error("There is no algorithm defined in the network.")
            raise SimulationException(SimulationException.ERRORS.ALGORITHM_NOT_FOUND)

        if self.algorithmState["finished"]:
            if len(self.algorithms) > self.algorithmState["index"] + 1:
                self.algorithmState["index"] += 1
                self.algorithmState["step"] = 1
                self.algorithmState["finished"] = False
            else:
                return None

        return self.algorithms[self.algorithmState["index"]]

    @property
    def algorithms(self):
        """
        Set algorithms by passing tuple of Algorithm subclasses.

        >>> sim.algorithms = (Algorithm1, Algorithm2,)

        For params pass tuples in form (Algorithm, params) like this

        >>> sim.algorithms = ((Algorithm1, {'param1': value,}), Algorithm2)

        """
        return self._algorithms

    @algorithms.setter
    def algorithms(self, algorithms: AlgorithmsParam):
        self.reset()
        self._algorithms = ()
        if not isinstance(algorithms, tuple):
            raise SimulationException(SimulationException.ERRORS.ALGORITHM)
        for algorithm in algorithms:
            if inspect.isclass(algorithm) and issubclass(algorithm, BaseAlgorithm):
                self._algorithms += (algorithm(self),)
            elif (
                isinstance(algorithm, tuple)
                and len(algorithm) == 2
                and issubclass(algorithm[0], BaseAlgorithm)
                and isinstance(algorithm[1], dict)
            ):
                self._algorithms += (algorithm[0](self, **algorithm[1]),)
            else:
                raise SimulationException(SimulationException.ERRORS.ALGORITHM)

        # If everything went ok, set algorithms param for coping
        self._algorithms_param = algorithms

    def get_dic(self):
        """
        Return all simulation data in the form of a dictionary.

        :return: A dictionary containing the simulation data.
        :rtype: dict
        """

        algorithms = {
            "%d %s" % (ind, alg.name): ("active" if alg == self.algorithms[self.algorithmState["index"]] else "")
            for ind, alg in enumerate(self.algorithms)
        }

        return {
            "algorithms": algorithms,  # A dictionary mapping algorithm names to their status (active or not).
            "algorithmState": {
                "name": (  # The name of the current algorithm.
                    self.get_current_algorithm().name if self.get_current_algorithm() else "No algorithm"
                ),
                "step": self.algorithmState["step"],  # The current step of the algorithm.
                "finished": self.algorithmState["finished"],  # Whether the algorithm has finished or not.
            },
        }



from collections.abc import Callable, Iterable
from itertools import product
from typing import TYPE_CHECKING, Any, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def pydistsim_equal_objects(obj1, obj2):
    """
    Compare two objects and their attributes, but allow for non immutable
    attributes to be equal up to their class.

    :param obj1: The first object to compare.
    :type obj1: object
    :param obj2: The second object to compare.
    :type obj2: object
    """
    classes = obj1.__class__ == obj2.__class__
    attr_names = attr_values = True
    if isinstance(obj1, object) and isinstance(obj2, object):
        attr_names = set(obj1.__dict__.keys()) == set(obj2.__dict__.keys())
    types = (str, tuple, int, int, bool, float, frozenset, bytes, complex)
    for key, value in list(obj1.__dict__.items()):
        other_value = getattr(obj2, key, None)
        if (isinstance(value, types) and value != other_value) or value.__class__ != other_value.__class__:
            attr_values = False
            break
    return classes and attr_names and attr_values


def with_typehint(baseclass: type[T]) -> type[T]:
    """
    Useful function to make mixins with baseclass typehint without actually inheriting from it.

    :param baseclass: The base class to use as a type hint.
    :type baseclass: type[T]
    :return: The base class if TYPE_CHECKING is True, otherwise object.
    :rtype: type[T]
    """
    if TYPE_CHECKING:
        return baseclass
    return object


def first(iterable: Iterable[T], default: U | None = None) -> T | U | None:
    """
    Return the first item in an iterable, or a default value if the iterable is empty.

    :param iterable: The iterable to get the first item from.
    :type iterable: Iterable[T]
    :param default: The default value to return if the iterable is empty.
    :type default: U | None

    :return: The first item in the iterable, or the default value if the iterable is empty.
    :rtype: T | U | None
    """

    iterator = iter(iterable)
    return next(iterator, default)


def measure_sortedness(sequence: Iterable[T], key: Callable[[T], Any] = None, reverse: bool = False) -> float:
    """
    Measure the sortedness of a sequence. A higher value means the sequence is more sorted.

    `sortedness = 1 - inversions / (n * (n - 1) / 2)`, so if `sortedness == 1` corresponds to a sorted sequence.


    :param sequence: The sequence to measure the sortedness of.
    :type sequence: Iterable[T]
    :param key: The key function to use to extract a comparison key from each element.
    :type key: Callable[[T], Any]
    :param reverse: Whether to sort the sequence in reverse order.
    :type reverse: bool
    :return: The sortedness of the sequence (between 0 and 1) and the inverted pairs found.
    :rtype: tuple[float, list[tuple[int, int]]]
    """

    if len(sequence) <= 1:
        return 1.0, []

    if not key:
        key = lambda x: x

    if reverse:
        key = lambda x: -key(x)

    sortedness = 0
    inverted_pairs = []
    for pair in product(range(len(sequence)), repeat=2):
        i, j = pair
        if i < j and key(sequence[i]) <= key(sequence[j]):
            sortedness += 1
        elif i != j:
            inverted_pairs.append(pair)

    return (sortedness / (len(sequence) * (len(sequence) - 1) / 2), inverted_pairs)


def len_is_one(iterable):
    """
    Returns true if the iterable has one element without consuming it entirely.
    """
    it = iter(iterable)
    try:
        next(it)
        return not next(it, False)
    except StopIteration:
        return False


def len_is_not_zero(iterable):
    """
    Returns true if the iterable has at least one element without consuming it entirely.
    """
    try:
        next(iter(iterable))
        return True
    except StopIteration:
        return False



from pydistsim.algorithm import NodeAlgorithm, StatusValues
from pydistsim.message import Message
from pydistsim.restrictions.communication import BidirectionalLinks
from pydistsim.restrictions.reliability import TotalReliability
from pydistsim.restrictions.topological import Connectivity, UniqueInitiator


class DFT(NodeAlgorithm):
    required_params = ()
    default_params = {"visitedAction": lambda node: None}

    class Status(StatusValues):
        INITIATOR = "INITIATOR"
        IDLE = "IDLE"
        DONE = "DONE"
        VISITED = "VISITED"

    S_init = (Status.INITIATOR, Status.IDLE)
    S_term = Status.DONE

    algorithm_restrictions = (
        BidirectionalLinks,
        TotalReliability,
        Connectivity,
        UniqueInitiator,
    )

    def initializer(self):
        for node in self.network.nodes():
            node.status = self.Status.IDLE
        ini_node = self.network.nodes_sorted()[0]
        ini_node.status = self.Status.INITIATOR
        ini_node.push_to_inbox(Message(meta_header=NodeAlgorithm.INI, destination=ini_node))

    @Status.INITIATOR
    def spontaneously(self, node, message):
        node.memory["parent"] = None
        node.memory["unvisited"] = list(node.neighbors())
        self.visitedAction(node)
        self.visit(node)

    @Status.INITIATOR
    def default(self, node, message):
        raise Exception("Dont disturb intitator!")

    @Status.IDLE
    def receiving(self, node, message):
        if message.header == "Token":
            node.memory["parent"] = message.source
            node.memory["unvisited"] = list(node.neighbors())
            node.memory["unvisited"].remove(node.memory["parent"])
            self.visitedAction(node)
            self.visit(node)
        else:
            raise Exception("Should not be here.")

    @Status.VISITED
    def receiving(self, node, message):
        if message.header == "Token":
            node.memory["unvisited"].remove(message.source)
            self.send_msg(node, Message(header="Backedge", destination=message.source))
        elif message.header == "Return":
            self.visit(node)
        elif message.header == "Backedge":
            self.visit(node)

    @Status.DONE
    def default(self, node, message):
        raise Exception("Im done!!!")

    def visit(self, node):
        if len(node.memory["unvisited"]) == 0:
            if node.memory["parent"] is not None:
                self.send_msg(node, Message(header="Return", destination=node.memory["parent"]))
            node.status = self.Status.DONE
        else:
            next_unvisited = node.memory["unvisited"].pop()
            self.send_msg(node, Message(header="Token", destination=next_unvisited))
            node.status = self.Status.VISITED


class DFStar(NodeAlgorithm):
    required_params = ()
    default_params = {"neighborsKey": "Neighbors", "visitedAction": lambda node: None}

    class Status(StatusValues):
        INITIATOR = "INITIATOR"
        IDLE = "IDLE"
        AVAILABLE = "AVAILABLE"
        VISITED = "VISITED"
        DONE = "DONE"

    S_init = (Status.INITIATOR, Status.IDLE)
    S_term = (Status.DONE,)

    algorithm_restrictions = (
        BidirectionalLinks,
        TotalReliability,
        Connectivity,
        UniqueInitiator,
    )

    def initializer(self):
        for node in self.network.nodes():
            node.status = self.Status.IDLE
        ini_node = self.network.nodes_sorted()[0]
        ini_node.status = self.Status.INITIATOR
        ini_node.push_to_inbox(Message(meta_header=NodeAlgorithm.INI, destination=ini_node))

    @Status.INITIATOR
    def spontaneously(self, node, message):
        node.memory["initiator"] = True
        node.memory["unvisited"] = list(node.neighbors())
        node.memory["next"] = node.memory["unvisited"].pop()
        if len(node.memory["unvisited"]) > 0:
            self.send_msg(
                node,
                Message(
                    header="T",
                    destination=node.memory["next"],
                ),
            )
            self.send_msg(node, Message(header="Visited", destination=node.memory["unvisited"]))
            node.status = self.Status.VISITED
            self.visitedAction(node)
        else:
            node.status = self.Status.DONE

    @Status.IDLE
    def receiving(self, node, message):
        if message.header == "T":
            node.memory["unvisited"] = list(node.neighbors())
            self.first_visit(node, message)

        if message.header == "Visited":
            node.memory["unvisited"] = list(node.neighbors())
            node.memory["unvisited"].remove(message.source)
            node.status = self.Status.AVAILABLE

    @Status.AVAILABLE
    def receiving(self, node, message):
        if message.header == "T":
            self.first_visit(node, message)

        if message.header == "Visited":
            node.memory["unvisited"].remove(message.source)

    @Status.VISITED
    def receiving(self, node, message):
        if message.header == "T":
            node.memory["unvisited"].remove(message.source)
            # late visited, should not happen in unitary communication delay
            if node.memory["next"] == message.source:
                self.visit(node, message)

        if message.header == "Visited":
            node.memory["unvisited"].remove(message.source)
            if node.memory["next"] == message.source:
                self.visit(node, message)

        if message.header == "Return":
            self.visit(node, message)

    @Status.DONE
    def default(self, node, message):
        pass

    def first_visit(self, node, message: Message):
        # TODO: initiator is redundant - it can be deduced from entry==None
        node.memory["initiator"] = False
        node.memory["entry"] = message.source
        self.visitedAction(node)
        try:
            node.memory["unvisited"].remove(message.source)
        except ValueError:
            pass
        if node.memory["unvisited"]:
            node.memory["next"] = node.memory["unvisited"].pop()
            self.send_msg(
                node,
                Message(
                    header="T",
                    destination=node.memory["next"],
                ),
            )

            self.send_msg(
                node,
                Message(
                    header="Visited",
                    destination=set(node.neighbors()) - {node.memory["entry"], node.memory["next"]},
                ),
            )
            node.status = self.Status.VISITED
        else:
            self.send_msg(node, Message(header="Return", destination=node.memory["entry"]))
            self.send_msg(
                node,
                Message(
                    header="Visited",
                    destination=set(node.neighbors()) - {node.memory["entry"]},
                ),
            )
            node.status = self.Status.DONE

    def visit(self, node, message):
        if node.memory["unvisited"]:
            node.memory["next"] = node.memory["unvisited"].pop()
            self.send_msg(
                node,
                Message(
                    header="T",
                    destination=node.memory["next"],
                ),
            )
        else:
            if not node.memory["initiator"]:
                self.send_msg(node, Message(header="Return", destination=node.memory["entry"]))
            node.status = self.Status.DONE



from collections import OrderedDict
from itertools import product
from typing import TYPE_CHECKING, Optional, TypeVar

from numpy import array, cos, log2, pi, sign, sin, sqrt
from numpy.random import rand

from pydistsim.logging import logger
from pydistsim.network.network import BidirectionalNetwork
from pydistsim.network.node import Node
from pydistsim.network.rangenetwork import BidirectionalRangeNetwork, RangeNetwork

if TYPE_CHECKING:
    from pydistsim.network.environment import Environment2D
    from pydistsim.network.network import NetworkType
    from pydistsim.network.rangenetwork import RangeNetworkType

T = TypeVar("T", bound="NetworkType")


class NetworkGenerator:
    """
    Class for generating networks with specified properties.


    :param n_count: int, number of nodes, if None, 100 is used
    :param n_min: int, minimum number of nodes, if not set it is equal to n_count
    :param n_max: int, maximum number of nodes, if not set it is equal to n_count
    :param enforce_connected: bool, if True network must be fully connected
    :param degree: int, average number of neighbors per node
    :param comm_range: int, nodes communication range, if None, 100 is used
        and it is a signal that this value can be changed if needed to
        satisfy other wanted properties (connected and degree)
    :param method: str, sufix of the name of the method used to generate network
    :param degree_tolerance: float, tolerance for degree parameter
    :param directed: bool, if True generated network is directed
    :param kwargs: network and node __init__ kwargs i.e.:
        - environment: Environment, environment in which the network should be created, if None
        Environment2D is used
        - rangeType: RangeType
        - algorithms: tuple
        - commRange: int, overrides `comm_range`
        - sensors: tuple

    Basic usage:

    >>> net_gen = NetworkGenerator()
    >>> net = net_gen.generate()

    """

    DIRECTED_NETWORK_T = RangeNetwork
    UNDIRECTED_NETWORK_T = BidirectionalRangeNetwork

    def __init__(
        self,
        n_count=None,
        n_min=None,
        n_max=None,
        enforce_connected=True,
        degree=None,
        comm_range=None,
        method="random_network",
        degree_tolerance=0.5,
        directed=False,
        **kwargs,
    ):
        self.n_count = n_count or 100
        self.n_min = self.n_count if n_min is None else n_min
        self.n_max = self.n_count if n_max is None else n_max
        if self.n_count < self.n_min or self.n_count > self.n_max:
            raise NetworkGeneratorException("Number of nodes (n_count parameter) must be between n_min and n_max.")
        if degree and degree >= self.n_max:
            raise NetworkGeneratorException(
                "Degree %d must be smaller than maximum number of nodes %d." % (degree, self.n_max)
            )
        # TODO: optimize recalculation of edges on bigger commRanges
        if degree:
            logger.warning("Generation could be slow for large degree parameter with bounded n_max.")
        self.enforce_connected = enforce_connected
        self.degree = degree
        self.degree_tolerance = degree_tolerance
        self.directed = directed
        self.comm_range = kwargs.pop("commRange", comm_range)
        # TODO: use subclass based generators instead of method based
        self.generate = self.__getattribute__("generate_" + method)
        self.kwargs = kwargs

    def _create_modify_network(self, net: Optional["RangeNetworkType"] = None, step=1) -> Optional["RangeNetworkType"]:
        """
        Helper method for creating a new network or modifying a given network.

        :param net: NetworkType object, optional
            The network to modify. If None, create a new network from scratch.
        :param step: int, optional
            If step > 0, the new network should be more dense. If step < 0, the new network should be less dense.

        :return: NetworkType object or None
            The modified network if successful, None otherwise.
        """
        if net is None:
            net_class = self.DIRECTED_NETWORK_T if self.directed else self.UNDIRECTED_NETWORK_T
            net = net_class(**self.kwargs)
            for _n in range(self.n_count):
                node = Node(commRange=self.comm_range, **self.kwargs)
                net.add_node(node)
        else:
            if step > 0:
                if len(net) < self.n_max:
                    node = Node(**self.kwargs)
                    net.add_node(node)
                    logger.trace("Added node, number of nodes: {}", len(net))
                elif not self.comm_range:
                    for node in net.nodes():
                        node.commRange += step
                    logger.trace("Increased commRange to {}", node.commRange)
                else:
                    return None
            else:
                min_node = net.nodes_sorted()[0]
                if len(net) > self.n_min and len(net) > 1:
                    net.remove_node(min_node)
                    logger.trace("Removed node, nodes left: {}", len(net))
                elif not self.comm_range:
                    for node in net:
                        node.commRange += step
                    logger.trace("Decreased commRange to {}", node.commRange)
                else:
                    return None
        return net

    def _are_conditions_satisfied(self, net: "RangeNetworkType"):
        """
        Check if the conditions for the network are satisfied.

        :param net: The network to check the conditions for.
        :type net: Network
        :return: The condition value.
        :rtype: int
        """
        cr = net.nodes_sorted()[0].commRange
        if self.enforce_connected and not net.is_connected():
            logger.debug("Not connected")
            return round(0.2 * cr)
        elif self.degree:
            diff = self.degree - net.avg_degree()
            if abs(diff) > self.degree_tolerance:
                logger.debug("Degree not satisfied: {} with {} nodes", net.avg_degree(), len(net))
                diff = sign(diff) * min(
                    max(abs(diff), 3), 7
                )  # If diff is too big, it will be set to 7, if it is too small, it will be set to 3
                condition_returned = round((sign(diff) * (round(diff)) ** 2) * cr / 100)
                logger.debug("Degree condition returned: {}", condition_returned)
                return condition_returned
        return 0

    def generate_random_network(
        self, net: Optional["RangeNetworkType"] = None, max_steps=1000
    ) -> Optional["RangeNetworkType"]:
        """
        Generates a random network with randomly positioned nodes.

        :param net: The network to modify. If not provided, a new network will be created.
        :type net: Optional[RangeNetworkType]
        :param max_steps: The maximum number of steps to take.
        :type max_steps: int
        :return: The generated network, optional.
        :rtype: Optional[RangeNetworkType]
        """
        # TODO: try some more advanced algorithm for situation when
        # both connected network and too small degree are needed
        # that is agnostic to actual dimensions of the environment
        steps = [0]
        while True:
            net = self._create_modify_network(net, steps[-1])
            if not net:
                break
            steps.append(self._are_conditions_satisfied(net))
            if len(steps) > max_steps:
                break
            if steps[-1] == 0:
                return net

        logger.error(
            "Could not generate connected network with given "
            "parameters. Try removing and/or modifying some of "
            "them."
        )

    def generate_neigborhood_network(self) -> "RangeNetworkType":
        """
        Generates a network where all nodes are in one hop neighborhood of
        at least one node.

        Finds out the node in the middle, which is the node with the minimum maximum
        distance to all other nodes, and sets that distance as the new commRange.

        This generator ignores all other parameters except comm_range and n counts.

        :return: The generated network.
        :rtype: RangeNetworkType
        """
        net = self._create_modify_network()

        max_distances = []
        for node in net:
            distances = [sqrt(sum((net.pos[node] - net.pos[neighbor]) ** 2)) for neighbor in net]
            max_distances.append(max(distances))
        min_distance = min(max_distances)
        for node in net:
            node.commRange = min_distance + 1
        return net

    def generate_homogeneous_network(self, randomness=0.11) -> Optional["RangeNetworkType"]:
        """
        Generates a network where nodes are located approximately homogeneous.

        :param randomness: Controls random perturbation of the nodes. It is given as a part of the environment size.
        :type randomness: float

        :return: The generated random network.
        :rtype: RangeNetworkType
        """
        net = self._create_modify_network()
        n = len(net)
        h, w = net.environment.image.shape
        assert net.environment.dim == 2  # works only for 2d environments
        size = w

        positions = generate_mesh_positions(net.environment, n)
        for i in range(n):
            pos = array([-1, -1])  # some non space point
            while not net.environment.is_space(pos):
                pos = positions[i, :n] + (rand(2) - 0.5) * (size * randomness)
            net.pos[net.nodes_sorted()[i]] = pos

        if isinstance(net, RangeNetwork):
            net.recalculate_edges()
        # TODO: this is not intuitive but generate_random network with net
        # given as argument will check if conditions are satisfied and act
        # accordingly, to change only commRange set limits to number of nodes
        self.n_count = self.n_max = self.n_min = n
        return self.generate_random_network(net)

    @staticmethod
    def __get_ring_pos(n: int, env: "Environment2D") -> list[tuple[float, float]]:
        """
        A helper method for generating positions of nodes on a circle.

        :param n: The number of nodes.
        :type n: int
        :param env: The environment object.
        :type env: Environment2D
        :return: The list of positions.
        :rtype: list[tuple[int, int]]
        """

        env_shape = env.image.shape
        center = (env_shape[0] // 2, env_shape[1] // 2)
        radius = 0.82 * (min(env_shape) / 2)  # 82% of the smallest dimension, leaves enough space for the node labels.

        nodes = []
        for i in range(n):
            angle = 2 * i * pi / n
            x = center[0] + radius * cos(angle)
            y = center[1] + radius * sin(angle)
            nodes.append((x, y))
        return nodes

    @classmethod
    def generate_complete_network(cls, n: int, network_type: type[T] = BidirectionalNetwork) -> T:
        """
        Generate a complete network with n nodes. The nodes are placed on a circle.

        :param n: The number of nodes in the complete network.
        :type n: int
        :return: The generated complete network.
        :rtype: BidirectionalNetwork
        """

        net = network_type()
        node_pos_list = cls.__get_ring_pos(n, net.environment)
        nodes = [net.add_node(pos=node_pos_list) for node_pos_list in node_pos_list]

        for i, j in product(range(n), range(n)):
            if i != j:
                net.add_edge(nodes[i], nodes[j])

        return net

    @classmethod
    def generate_ring_network(cls, n: int, network_type: type[T] = BidirectionalNetwork) -> T:
        """
        Generate a ring network with n nodes. The nodes are placed on a circle.
        If network_type is a directed type, the links between the nodes are one-way only so the network is fully
        connected.

        :param n: The number of nodes in the ring network.
        :type n: int
        :param network_type: The type of network to generate.
        :type network_type: NetworkType
        :return: The generated ring network. Same type as the network_type parameter.
        :rtype: NetworkType
        """

        net = network_type()
        node_pos_list = cls.__get_ring_pos(n, net.environment)
        nodes = [net.add_node(pos=node_pos_list) for node_pos_list in node_pos_list]

        for i in range(n):
            if i != (i + 1) % n:
                net.add_edge(nodes[i], nodes[(i + 1) % n])

        return net

    @classmethod
    def generate_star_network(cls, n: int, network_type: type[T] = BidirectionalNetwork) -> T:
        """
        Generate a star network with n nodes. The nodes are placed on a circle with one node in the center.
        If network_type is a directed type, the links are only from the center node to the other nodes.

        :param n: The number of TOTAL nodes in the star network.
        :type n: int
        :param network_type: The type of network to generate.
        :type network_type: NetworkType
        :return: The generated star network.
        :rtype: NetworkType
        """

        if n <= 1:
            raise ValueError("Star network requires at least 2 nodes.")

        net = network_type()
        node_pos_list = cls.__get_ring_pos(n - 1, net.environment)

        center = (
            net.environment.image.shape[0] / 2,
            net.environment.image.shape[1] / 2,
        )

        center_node = net.add_node(pos=center)

        for i in range(n - 1):
            node = net.add_node(pos=node_pos_list[i])
            net.add_edge(center_node, node)

        return net

    @staticmethod
    def generate_hypercube_network(
        n: int | None = None,
        dimension: int | None = None,
        use_binary_labels: bool = True,
        network_type: type[T] = BidirectionalNetwork,
    ) -> T:
        """
        Generate a hypercube network of the given dimension (or node count). The nodes are placed in a hypercube
        structure.

        :param n: The number of nodes in the hypercube. If None, the dimension parameter must be set.
        :type n: int
        :param dimension: The dimension of the hypercube. If None, the n parameter must be set.
        :type dimension: int
        :param network_type: The type of network to generate.
        :type network_type: NetworkType
        :return: The generated hypercube network.
        :rtype: NetworkType
        """
        if n is None and dimension is None:
            raise ValueError("Either the number of nodes or the dimension of the hypercube must be set.")
        if n is not None and dimension is not None:
            raise ValueError("Only one of the number of nodes or the dimension of the hypercube can be set.")
        if n is not None:
            dimension = int(log2(n))
            if 2**dimension != n:
                raise ValueError("The number of nodes must be a power of 2.")

        LABEL_KEY = "HYPERCUBE_LABEL"

        def create_hypercube(dim):
            # Recursive function to create the hypercube
            # A hypercube of dimension n is the disjoint union of two hypercubes of dimension n-1, forming a perfect
            # matching between the nodes of the two hypercubes.
            node_pos = OrderedDict()
            edges = []
            dist = 1

            if dim == 0:
                node = Node()
                node.memory[LABEL_KEY] = "0"
                node_pos[node] = (0, 0)
            elif dim == 1:
                node1 = Node()
                node1.memory[LABEL_KEY] = "0"
                node2 = Node()
                node2.memory[LABEL_KEY] = "1"
                node_pos[node1] = (0, 0)
                node_pos[node2] = (dist, 0)
                edges.append((node1, node2))
            else:
                part_1_nodes, part_1_edges = create_hypercube(dim - 1)
                part_2_nodes, part_2_edges = create_hypercube(dim - 1)

                edges += part_1_edges
                edges += part_2_edges

                part_1_width = max([pos[0] for pos in part_1_nodes.values()])
                part_1_height = max([pos[1] for pos in part_1_nodes.values()])

                if dim == 1:  # Grow only in the x-axis
                    x_offset = dist
                    y_offset = 0
                elif dim == 2:  # Grow only in the y-axis
                    x_offset = 0
                    y_offset = dist
                elif dim == 3:  # Grow equally in both axes
                    x_offset = dist / 2
                    y_offset = dist / 2
                elif dim % 2 == 0:  # Even dimensions grow in the x-axis
                    x_offset = part_1_width * 1.1
                    y_offset = dist / 4
                else:  # Odd dimensions grow in the y-axis
                    x_offset = dist / 4
                    y_offset = part_1_height * 1.1

                for node, pos in part_1_nodes.items():
                    node_pos[node] = pos
                    node.memory[LABEL_KEY] = "0" + node.memory[LABEL_KEY]
                for node, pos in part_2_nodes.items():
                    node_pos[node] = (pos[0] + x_offset, pos[1] + y_offset)
                    node.memory[LABEL_KEY] = "1" + node.memory[LABEL_KEY]

                for node1, node2 in zip(part_1_nodes.keys(), part_2_nodes.keys()):
                    edges.append((node1, node2))

            return node_pos, edges

        # Create the nodes
        node_pos_list, edges = create_hypercube(dimension)

        # Create the network
        net = network_type()
        canvas_shape = net.environment.image.shape
        usable_shape = (canvas_shape[1] * 0.82, canvas_shape[0] * 0.82)

        # Normalize the positions
        max_x = max(max([pos[0] for pos in node_pos_list.values()]), 1)
        max_y = max(max([pos[1] for pos in node_pos_list.values()]), 1)

        for node, pos in node_pos_list.items():
            # If the dimension is 0 or 1, the nodes are in the center
            x = pos[0] if dimension != 0 else 0.5
            y = pos[1] if dimension not in (0, 1) else 0.5
            final_pos = (x / max_x * usable_shape[0], y / max_y * usable_shape[1])

            # Add the node
            net.add_node(node, pos=final_pos)
            if use_binary_labels:
                net.labels[node] = node.memory[LABEL_KEY]

            # Remove the label from the memory as it represents some knowledge of the network
            del node.memory[LABEL_KEY]

        # Add the edges to the network
        for node1, node2 in edges:
            net.add_edge(node1, node2)
            if net.is_directed():
                net.add_edge(node2, node1)

        return net

    @classmethod
    def generate_mesh_network(
        cls,
        n: int = None,
        a: int = None,
        b: int = None,
        torus: bool = False,
        network_type: type[T] = BidirectionalNetwork,
    ) -> T:
        """
        Generate a mesh (or torus) network with n nodes (or a x b nodes).
        If network_type is a directed type, the links are only north and east. Only with torus, this gives a fully
        connected network nonetheless.

        :param n: The number of nodes in the mesh/torus network. Optional.
        :type n: int
        :param a: Width of the mesh/torus. Optional.
        :type a: int
        :param b: Height of the mesh/torus. Optional.
        :type b: int
        :param torus: Whether or not to add "returning" edges.
        :type torus: bool
        :param network_type: The type of network to generate.
        :type network_type: NetworkType
        :return: The generated mesh/torus network.
        :rtype: NetworkType
        """

        if a is None or b is None:
            if n is None:
                raise ValueError("Mesh network generator needs (a, b) or n.")
            if int(sqrt(n)) != sqrt(n):
                raise ValueError("Value n given must be a perfect square.")

            a = b = int(sqrt(n))

        node_pos_list = {(i, j): Node() for i in range(a) for j in range(b)}

        net = network_type()
        canvas_shape = net.environment.image.shape
        usable_shape = (canvas_shape[1] * 0.82, canvas_shape[0] * 0.82)
        bottom_margin = (
            canvas_shape[1] * (1 - 0.83) / 4,
            canvas_shape[0] * (1 - 0.83) / 4,
        )

        # Normalize the positions
        max_x = max(max([pos[0] for pos in node_pos_list.keys()]), 1)
        max_y = max(max([pos[1] for pos in node_pos_list.keys()]), 1)

        for (x, y), node in node_pos_list.items():
            # If the dimension is 0 or 1, the nodes are in the center
            final_pos = (
                x / max_x * usable_shape[0] + bottom_margin[0],
                y / max_y * usable_shape[1] + bottom_margin[1],
            )
            net.add_node(node, pos=final_pos)

        for (x, y), node in node_pos_list.items():
            north_and_east_neighbors = [(x, y + 1), (x + 1, y)] if not torus else [(x, (y + 1) % b), ((x + 1) % a, y)]
            for pos_neigh in north_and_east_neighbors:
                if pos_neigh in node_pos_list:
                    net.add_edge(node, node_pos_list[pos_neigh])

        return net

    generate_grid_network = generate_mesh_network


def generate_mesh_positions(env, n):
    """
    Generate mesh positions for the given environment and number of intersections.

    :param env: The environment object.
    :type env: Environment
    :param n: The desired number of intersections.
    :type n: int
    :return: An array of mesh positions.
    :rtype: numpy.ndarray
    """
    h, w = env.image.shape
    # initial d
    d = sqrt(h * w / n)

    def get_mesh_pos(d, dx, dy, w, h):
        return [
            (xi_yi[0] * d + dx, xi_yi[1] * d + dy)
            for xi_yi in product(list(range(int(round(w / d)))), list(range(int(round(h / d)))))
        ]

    n_mesh = 0
    direction = []
    while True:
        n_mesh = len([1 for pos in get_mesh_pos(d, 0.5 * d, 0.5 * d, w, h) if env.is_space(pos)])
        direction.append(sign(n - n_mesh))
        if n_mesh == n or (len(direction) >= 10 and abs(sum(direction[-3:])) < 3 and n_mesh > n):
            break
        d *= sqrt(n_mesh / float(n))
    return array(tuple(pos for pos in get_mesh_pos(d, 0.5 * d, 0.5 * d, w, h) if env.is_space(pos)))
    # TODO: n_mesh could be brought closer to n with modification of dx and dy
    # dx = 0.5*d
    # dy = 0.5*d


class NetworkGeneratorException(Exception):
    pass



import unittest

from numpy import array, concatenate, dot, nan
from numpy.ma.core import cos, sin
from numpy.random import rand
from scipy.constants import pi

from pydistsim.utils.localization.aoastitcher import AoAStitcher
from pydistsim.utils.localization.diststitcher import DistStitcher
from pydistsim.utils.localization.helpers import get_rms
from pydistsim.utils.memory.positions import Positions


class TestStitchers(unittest.TestCase):

    def setUp(self):
        # stitch_test.cdr
        self.test_cluster = {
            0: array([0.0, 0.0, nan]),
            1: array([3.0, -2.0, nan]),
            2: array([6.0, 2.0, nan]),
            3: array([2.0, 5.0, nan]),
            4: array([-2.0, 4.0, nan]),
            5: array([2.0, -5.0, nan]),
            6: array([7.0, -3.0, nan]),
        }

    def test_aoa_stitcher(self):
        src = self.setup_positions([[0, 1, 2, 3, 4]])
        dst = self.setup_positions([[0, 1, 2, 5, 6]])

        # theta = 60/180.*pi
        # R = array([[cos(theta), -sin(theta)], [sin(theta), cos(theta)]])  #ccw
        # self.transform(src, R, [-1.62, -0.8], 0.5)  #transform src positions
        self.transform(src)

        stitcher = AoAStitcher()
        stitcher.stitch(dst, src)

        rms = get_rms(self.test_cluster, dst)
        self.assertTrue(rms < 0.001, "Stitching was not good enough.")
        self.assertTrue(len(dst) == 1, "Wrong number of subclusters.")

    def test_dist_stitcher(self):
        src = self.setup_positions([[0, 1, 2, 3, 4]])
        dst = self.setup_positions([[0, 1, 2, 5, 6]])

        self.transform(src, s_i=1)

        stitcher = DistStitcher()
        stitcher.stitch(dst, src)

        self.assertTrue(len(dst) == 1, "Wrong number of subclusters.")
        rms = get_rms(self.test_cluster, dst)
        self.assertTrue(rms < 0.001, "Stitching was not good enough.")

    def test_dist_stitcher2(self):
        """Test dist stitch with only 2 common nodes - not enough."""
        src = self.setup_positions([[0, 1, 3, 4]])
        dst = self.setup_positions([[0, 1, 2, 5, 6]])
        self.transform(src, s_i=1)  # transform src positions

        stitcher = DistStitcher()
        stitcher.stitch(dst, src)

        self.assertTrue(len(dst) == 2, "Wrong number of subclusters.")
        rms = get_rms(self.test_cluster, dst)
        self.assertFalse(rms < 0.001, "Stitching was too good.")
        # with align rms should be ok
        rms_align = get_rms(self.test_cluster, dst, align=True)
        self.assertTrue(rms_align < 0.001, "Stitching was not good enough.")

    def test_dist_stitcher3(self):
        """Test dist stitch with only 1 common node - not enough."""
        src = self.setup_positions([[0, 3, 4]])
        dst = self.setup_positions([[0, 1, 2, 5, 6]])
        self.transform(src, s_i=1)  # transform src positions

        stitcher = DistStitcher()
        stitcher.stitch(dst, src)

        self.assertTrue(len(dst) == 2, "Wrong number of subclusters.")
        # without align new subcluster should be wrongly (not) rotated
        rms = get_rms(self.test_cluster, dst)
        self.assertFalse(rms < 0.001, "Stitching was too good.")
        # with align rms should be ok
        rms_align = get_rms(self.test_cluster, dst, align=True)
        self.assertTrue(rms_align < 0.001, "Stitching was not good enough.")

    def test_dist_intrastitch(self):
        """
        Test dist intrastitch

        Clusters are setup so that first stitch is src[0] with dst[1] and
        then dst[0] and dst[1] can be intrastitched.

        """
        src = self.setup_positions([[1, 2, 4, 5, 6]])
        dst = self.setup_positions([[0, 2, 3, 4], [0, 1, 2, 6]])
        self.transform(src, s_i=1)  # transform src positions
        self.transform(dst, s_i=1)  # transform src positions

        stitcher = DistStitcher()
        stitcher.stitch(dst, src)

        self.assertTrue(len(dst) == 1, "Wrong number of subclusters.")
        # without align new subcluster should be wrongly (not) rotated
        rms = get_rms(self.test_cluster, dst)
        self.assertFalse(rms < 0.001, "Stitching was too good.")
        # with align rms should be ok
        rms_align = get_rms(self.test_cluster, dst, align=True)
        self.assertTrue(rms_align < 0.001, "Stitching was not good enough.")

    def setup_positions(self, subclusters):
        """Return new cluster with nodes form main_cluster and with
        subclusters defined as in subcluster argument."""

        positions = [{k: v for k, v in list(self.test_cluster.items()) if k in sub} for sub in subclusters]
        return Positions(positions)

    def transform(self, positions, R_i=None, t_i=None, s_i=None, flip=False):
        """
        Return subclusters with (randomly) rotated translated and scaled
        positions. If R_i, s_i or t_i is given then that part of transformation
        is not random.

        """

        for sub in positions:
            t = t_i or rand(2) * 10
            s = s_i or rand() * 2
            if R_i is None:
                th = 2 * pi * rand()
                # ccw
                R = array([[cos(th), -sin(th)], [sin(th), cos(th)]])
            else:
                R = R_i
            if flip:
                # TODO: make R with flip
                pass

            for node, pos in list(sub.items()):
                sub[node] = concatenate((dot(dot(s, R), pos[:2]) + t, [nan]))



"""
Sensors provide a way for node to interact with its environment.

Sensors can also be used to satisfy algorithm prerequisites.

Generally sensors should incorporate some model of measurement insecurity that
is inherent in real world sensors. This is implemented as a
:class:`ProbabilityFunction`.

Basic usage:

>>> node.compositeSensor = ('DistSensor','AoASensor')
>>> node.compositeSensor.sensors
(<pydistsim.network.sensor.DistSensor at 0x6d3fbb0>,
 <pydistsim.network.sensor.AoASensor at 0x6d3f950>)

To manually set sensor parameters first make an sensor instance:

>>> import scipy.stats
>>> aoa_sensor = AoASensor({'pf': scipy.stats.norm, 'scale': 10*pi/180 })
>>> node.compositeSensor = (aoa_sensor,)

"""

import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING

from numpy import arctan2, pi, sqrt
from scipy.stats import rv_continuous, rv_discrete

if TYPE_CHECKING:
    from pydistsim.network import Node


class Sensor(ABC):
    """
    Abstract base class for all Sensors.

    Sensor provides a certain capability for a node, providing information about
    the outside world. It could be a capability to detect neighbors, measure
    distance to them, or retrieve the environment temperature.

    :param dict_args: A dictionary containing the scale and probability function.
    :type dict_args: dict
    :param scale: The scale parameter for the probability function.
    :type scale: float
    :param pf: The probability function (e.g. :py:data:`scipy.stats.norm`).
    :type pf: rv_continuous or rv_discrete
    """

    def __init__(self, dict_args=None, /, scale=None, pf: rv_continuous | rv_discrete = None):
        if dict_args:
            scale = dict_args.get("scale")
            pf = dict_args.get("pf")

        if pf and scale:
            self.probabilityFunction = ProbabilityFunction(scale, pf)
        else:
            self.probabilityFunction = None

    @classmethod
    def name(cls):
        """
        Get the name of the Sensor class.

        :return: The name of the Sensor class.
        :rtype: str
        """
        return cls.__name__

    @abstractmethod
    def read(self) -> dict:
        """
        Read the sensor data.

        This method should be overridden in a subclass.

        :return: The sensor data.
        :rtype: dict
        """
        pass


def node_in_network(fun: Callable):
    """Decorator function that checks if node is in network.

    :param fun: The function to be decorated.
    :type fun: Callable
    :return: The decorated function.
    :rtype: Callable
    """

    @wraps(fun)
    def f(sensor: Sensor, node: "Node"):
        if not node.network:
            raise SensorError("Cannot take a sensor reading if node is outside of a network.")
        return fun(sensor, node)

    return f


class AoASensor(Sensor):
    """
    Provides azimuth between node and its neighbors.

    This sensor calculates the azimuth angle between a node and its neighbors in a network.
    It uses the position and orientation information of the nodes to calculate the azimuth angle.
    """

    @node_in_network
    def read(self, node: "Node"):
        """
        Reads the azimuth angle between a node and its neighbors.

        :param node: The node for which to calculate the azimuth angle.
        :type node: Node
        :return: A dictionary containing the azimuth angle measurements between the node and its neighbors.
        :rtype: dict
        """
        network = node.network
        measurements = {}
        p = network.pos[node]
        o = network.ori[node]
        for neighbor in network.neighbors(node):
            v = network.pos[neighbor] - p
            measurement = (arctan2(v[1], v[0]) - o) % (2 * pi)
            measurement = self.probabilityFunction.getNoisyReading(measurement)
            measurements.update({neighbor: measurement})
        return {"AoA": measurements}


class DistSensor(Sensor):
    """Provides distance between node and its neighbors."""

    @node_in_network
    def read(self, node: "Node"):
        """
        Read the distances from the current node to its neighbors.

        :param node: The current node.
        :type node: Node
        :return: A dictionary containing the distances to the neighbors.
        :rtype: dict
        """
        network = node.network
        measurements = {}
        p = network.pos[node]
        for neighbor in network.neighbors(node):
            pn = network.pos[neighbor]
            measurement = sqrt(sum(pow(p - pn, 2)))
            measurement = self.probabilityFunction.getNoisyReading(measurement)
            measurements.update({neighbor: measurement})
        return {"Dist": measurements}


class TruePosSensor(Sensor):
    """Provides node's true position."""

    @node_in_network
    def read(self, node: "Node"):
        """
        Read the sensor data from the given node.

        :param node: The node from which to read the sensor data.
        :type node: Node

        :return: A dictionary containing the sensor data.
        :rtype: dict
        """
        return {"TruePos": node.network.pos[node]}


class CompositeSensor:
    """
    Wrap multiple sensors, coalesce results and return composite readout.

    This class is not a sensor itself, i.e. subclass of :class:`Sensor`,
    instead it serves as a placeholder for multiple sensors that can be
    attached to a :class:`Node`.


    :param node: The Node that has this composite sensor attached to.
    :type node: Node
    :param componentSensors: Tuple of Sensor subclasses or their class names.
    :type componentSensors: tuple[type[Sensor] | str]
    """

    def __init__(self, node: "Node", componentSensors: tuple[type[Sensor] | str] | None = None):
        self.node = node
        self._sensors = ()
        self.sensors = componentSensors or ()

    @property
    def sensors(self) -> tuple[Sensor]:
        """
        Get the sensors associated with the object.

        :return: A tuple of Sensor objects.
        """
        return self._sensors

    @sensors.setter
    def sensors(self, sensors: tuple[type[Sensor] | str]):
        self._sensors: tuple[Sensor] = ()
        # instantiate sensors passed by class name
        for cls in Sensor.__subclasses__():
            if cls.__name__ in sensors:
                self._sensors += (cls(),)
        # instantiate sensors passed by class
        for cls in sensors:
            if inspect.isclass(cls) and issubclass(cls, Sensor):
                self._sensors += (cls(),)
        # add sensors that are already instantiated
        for sensor in sensors:
            if isinstance(sensor, Sensor):
                self._sensors += (sensor,)

    def get_sensor(self, name: str) -> Sensor:
        """
        Get a sensor by its name.

        :param name: The name of the sensor.
        :type name: str
        :return: The sensor object.
        :rtype: Sensor
        :raises SensorError: If multiple or no sensors are found with the given name.
        """
        sensor = [s for s in self._sensors if s.name() == name]
        if len(sensor) != 1:
            raise SensorError("Multiple or no sensors found with name %s" % name)
        return sensor[0]

    def read(self):
        """
        Read measurements from all sensors.

        :return: A dictionary containing the measurements from all sensors.
        :rtype: dict
        """
        measurements = {}
        for sensor in self._sensors:
            measurements.update(sensor.read(self.node))
        return measurements


class ProbabilityFunction:
    """
    Provides a way to get noisy reading.

    :param scale: The scale parameter for the probability function.
    :type scale: float

    :param pf: The probability function (e.g. :py:data:`scipy.stats.norm`).
    :type pf: rv_continuous or rv_discrete
    """

    def __init__(self, scale, pf: rv_continuous | rv_discrete):
        self.pf = pf  # class or gen object
        self.name = self.pf.__class__.__name__
        self.scale = scale

    def getNoisyReading(self, value):
        return self.pf.rvs(scale=self.scale, loc=value)


class SensorError(Exception):
    pass



"""
This module contains restrictions related to time and how it is handled in the simulation.
"""

from abc import ABC
from typing import TYPE_CHECKING

from pydistsim.message import Message, MetaHeader
from pydistsim.restrictions.base_restriction import (
    ApplicableRestriction,
    CheckableRestriction,
)
from pydistsim.utils.helpers import len_is_not_zero

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


class TimeRestriction(CheckableRestriction, ABC):
    """
    Base class for restrictions relating to time.

    In fact, the general model makes no assumption about delays (except that they are finite).
    """


class BoundedCommunicationDelays(TimeRestriction):
    """
    There exists a constant T such that, in the absence of failures, the communication delay of any message on any link
    is at most T.
    """

    help_message = (
        "Delays are not upper bounded. Choose a value for `network.behavioral_properties` such that "
        "`bounded_communication_delays` is True.\n"
        "The property NetworkBehaviorModel.RandomDelayCommunication should be a good example."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return network.behavioral_properties.bounded_communication_delays


class UnitaryCommunicationDelays(TimeRestriction):
    """
    In the absence of failures, the communication delay of any message on any link is one unit of time.
    """

    help_message = (
        "Delays are not unitary. Choose a value for `network.behavioral_properties` such that "
        "`message_delay_indicator` is `None`.\n"
        "The property NetworkBehaviorModel.UnorderedCommunication should be a good example."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return network.behavioral_properties.message_delay_indicator is None


class SynchronizedClocks(TimeRestriction):
    """
    All local clocks are incremented by one unit simultaneously and the interval of time between successive increments
    is constant.
    """

    help_message = (
        "Clocks are not synchronized. Choose a value for `network.behavioral_properties` such that "
        "`clock_increment` is `None`.\n"
        "The property NetworkBehaviorModel.IdealCommunication should be a good example."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return network.behavioral_properties.clock_increment is None


class SimultaneousStart(TimeRestriction, ApplicableRestriction):
    """
    All nodes start the algorithm at the same time.

    Note that "at the same time" is a given under this simulation model but that does not mean that the nodes are
    synchronized, nor that their messages arrive with the same delay.
    """

    help_message = (
        "Not every node starts at the beginning. Try inserting a`INITIALIZATION_MESSAGE` in the inbox of every node.\n"
        + ApplicableRestriction.help_message
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        def message_is_ini(message: "Message"):
            return message.meta_header == MetaHeader.INITIALIZATION_MESSAGE

        nodes_with_ini = (node for node in network if len_is_not_zero(filter(message_is_ini, node.inbox)))
        return len(list(nodes_with_ini)) == len(network)

    @classmethod
    def apply(self, network: "NetworkType") -> None:
        for node in network.nodes():
            node.push_to_inbox(Message(meta_header=MetaHeader.INITIALIZATION_MESSAGE, destination=node))



# flake8: noqa: E731

from pydistsim.message import Message
from pydistsim.network import NetworkGenerator
from pydistsim.network.behavior import NetworkBehaviorModel
from pydistsim.network.network import Network
from pydistsim.restrictions.communication import (
    BidirectionalLinks,
    MessageOrdering,
    ReciprocalCommunication,
)
from pydistsim.restrictions.knowledge import InitialDistinctValues, NetworkSize
from pydistsim.restrictions.reliability import TotalReliability
from pydistsim.restrictions.time import (
    BoundedCommunicationDelays,
    SimultaneousStart,
    SynchronizedClocks,
    UnitaryCommunicationDelays,
)
from pydistsim.restrictions.topological import (
    CompleteGraph,
    Connectivity,
    CycleGraph,
    StarGraph,
    TreeGraph,
    UniqueInitiator,
)
from pydistsim.utils.testing import PyDistSimTestCase


class TestRestrictions(PyDistSimTestCase):

    def test_TotalReliability(self):
        net = NetworkGenerator(10).generate_random_network()

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=None,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=True,
        )

        net.behavioral_properties = behavior
        assert TotalReliability.check(net)

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=False,
        )

        net.behavioral_properties = behavior
        assert not TotalReliability.check(net)

    def test_MessageOrdering(self):
        net = NetworkGenerator(10).generate_random_network()

        behavior = NetworkBehaviorModel(
            message_ordering=True,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=False,
        )

        net.behavioral_properties = behavior
        assert MessageOrdering.check(net)

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=False,
        )

        net.behavioral_properties = behavior
        assert not MessageOrdering.check(net)

    def test_ReciprocalCommunication(self):
        net = Network()  # directed graph
        n1, n2, n3 = net.add_node(), net.add_node(), net.add_node()

        net.add_edge(n1, n2)
        assert not ReciprocalCommunication.check(net)
        net.add_edge(n2, n1)
        assert ReciprocalCommunication.check(net)

        net.add_edge(n3, n2)
        assert not ReciprocalCommunication.check(net)
        net.add_edge(n3, n1)
        assert not ReciprocalCommunication.check(net)
        net.add_edge(n1, n3)
        assert not ReciprocalCommunication.check(net)
        net.add_edge(n2, n3)

        print(net.nodes_sorted())
        print([e for e in net.edges()])
        assert ReciprocalCommunication.check(net)

        und_net = NetworkGenerator(10, directed=False).generate_random_network()
        assert ReciprocalCommunication.check(und_net)

    def test_BidirectionalLinks(self):
        net = NetworkGenerator(10, directed=False).generate_random_network()
        assert BidirectionalLinks.check(net)

        net = NetworkGenerator(10, directed=True).generate_random_network()
        assert not BidirectionalLinks.check(net)

    def test_InitialDistinctValues(self):
        net = NetworkGenerator(10).generate_random_network()
        InitialDistinctValues.apply(net)
        InitialDistinctValues.check(net)

        for node in net.nodes():
            node.memory[InitialDistinctValues.KEY] = 1
        assert not InitialDistinctValues.check(net)

    def test_NetworkSize(self):
        net = NetworkGenerator(10).generate_random_network()
        NetworkSize.apply(net)
        assert NetworkSize.check(net)

        for node in net.nodes():
            node.memory[NetworkSize.KEY] = 9
        assert not NetworkSize.check(net)

    def test_SynchronizedClocks(self):
        net = NetworkGenerator(10).generate_random_network()

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=None,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=False,
        )

        net.behavioral_properties = behavior
        assert SynchronizedClocks.check(net)

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=False,
        )

        net.behavioral_properties = behavior
        assert not SynchronizedClocks.check(net)

    def test_BoundedCommunicationDelays(self):
        net = NetworkGenerator(10).generate_random_network()

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: 1000,
            bounded_communication_delays=True,
        )

        net.behavioral_properties = behavior
        assert BoundedCommunicationDelays.check(net)

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: float("inf"),
            bounded_communication_delays=False,
        )

        net.behavioral_properties = behavior
        assert not BoundedCommunicationDelays.check(net)

    def test_UnitaryCommunicationDelays(self):
        net = NetworkGenerator(10).generate_random_network()

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=None,
            bounded_communication_delays=True,
        )

        net.behavioral_properties = behavior
        assert UnitaryCommunicationDelays.check(net)

        behavior = NetworkBehaviorModel(
            message_ordering=False,
            message_loss_indicator=lambda network, message: True,
            clock_increment=lambda node: 100,
            message_delay_indicator=lambda network, message: float("inf"),
            bounded_communication_delays=True,
        )

        net.behavioral_properties = behavior
        assert not UnitaryCommunicationDelays.check(net)

    def test_SimultaneousStart(self):
        net = NetworkGenerator(10).generate_random_network()
        assert not SimultaneousStart.check(net)

        for node in net.nodes():
            node.push_to_inbox(Message(meta_header=Message.META_HEADERS.INITIALIZATION_MESSAGE))

        assert SimultaneousStart.check(net)

    def test_Connectivity(self):
        net = NetworkGenerator(10, enforce_connected=True).generate_random_network()
        assert Connectivity.check(net)

        net = Network()
        net.add_node()
        net.add_node()
        assert not Connectivity.check(net)

    def test_UniqueInitiator(self):
        net = NetworkGenerator(10).generate_random_network()

        for node in net.nodes_sorted()[0:2]:
            node.push_to_inbox(Message(meta_header=Message.META_HEADERS.INITIALIZATION_MESSAGE))

        assert not UniqueInitiator.check(net)

        net = NetworkGenerator(10).generate_random_network()

        node = net.nodes_sorted()[0]
        node.push_to_inbox(Message(meta_header=Message.META_HEADERS.INITIALIZATION_MESSAGE))

        assert UniqueInitiator.check(net)

    def test_CompleteGraph(self):
        net = Network()

        nodes = [net.add_node() for _ in range(10)]
        for i, n1 in enumerate(nodes):
            for j, n2 in enumerate(nodes):
                if i != j:
                    net.add_edge(n1, n2)

        assert CompleteGraph.check(net)

        net = NetworkGenerator(10).generate_random_network()
        assert not CompleteGraph.check(net)

    def test_CycleGraph(self):
        net = Network()
        LEN = 10
        nodes = [net.add_node() for _ in range(LEN)]
        for i, n1 in enumerate(nodes):
            n2 = nodes[(i + 1) % LEN]
            net.add_edge(n1, n2)

        assert CycleGraph.check(net)

        net = Network()
        nodes = [net.add_node() for _ in range(LEN)]
        for i, n1 in enumerate(nodes[:-2]):
            n2 = nodes[(i + 1) % LEN]
            net.add_edge(n1, n2)

        assert not CycleGraph.check(net)

    def test_TreeGraph(self):
        net = Network()
        n1, n2, n3, n4, n5 = net.add_node(), net.add_node(), net.add_node(), net.add_node(), net.add_node()

        net.add_edge(n1, n2), net.add_edge(n1, n3), net.add_edge(n2, n4), net.add_edge(n2, n5)
        assert TreeGraph.check(net)

        net.add_edge(n3, n4)  # cycle
        assert not TreeGraph.check(net)

    def test_StarGraph(self):
        net = Network()
        n1, n2, n3, n4, n5 = net.add_node(), net.add_node(), net.add_node(), net.add_node(), net.add_node()

        net.add_edge(n1, n2), net.add_edge(n1, n3), net.add_edge(n1, n4)
        assert not StarGraph.check(net)

        net.add_edge(n1, n5)  # missing edge
        assert StarGraph.check(net)

        net.add_edge(n2, n3)  # cycle
        assert not StarGraph.check(net)



"""
Logger configuration for the package.
"""

import sys
from enum import StrEnum

import loguru

logger = loguru.logger
"Loguru logger instance for the package."


class LogLevels(StrEnum):
    """
    Enum class representing different log levels.
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class LevelFilter:
    def __init__(self, level: LogLevels):
        self.level = level

    def __call__(self, record):
        level_no = logger.level(self.level).no
        return record["level"].no >= level_no


main_filter = LevelFilter(LogLevels.WARNING)

logger.remove()

logger.add(sys.stdout, filter=main_filter, level=0)

logger.disable("pydistsim")


def set_log_level(level: LogLevels):
    main_filter.level = level


def enable_logger():
    logger.enable("pydistsim")


def disable_logger():
    logger.disable("pydistsim")
